/*
 *  Dynamic filter parser.
 *
 *  Based on vf.c
 *
 *  Copyright (C) 2007 Attila Ötvös <attila@onebithq.com>
 *
 *  MPlayer is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  MPlayer is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with MPlayer; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>
#include <ctype.h>

#include "config.h"

#include "mp_msg.h"
#include "help_mp.h"

#include "m_option.h"
#include "m_struct.h"

#include "dynamicfilter.h"

#define PARSE_BUFFER_SIZE        8192

extern const dfcmd_info_t vfcmd_info_import;
extern const dfcmd_info_t afcmd_info_import;

// list of available filters:
static const dfcmd_info_t* cmd_video_filter_list[]={
    &vfcmd_info_import,
    NULL
};

static const dfcmd_info_t* cmd_audio_filter_list[]={
    &afcmd_info_import,
    NULL
};

typedef struct filter_param_opts_s {
    int    id;
    int    av;
    char  *filtername;
    char  *cmd;
    float  pts;
} filter_param_opts_t;

static filter_param_opts_t const filter_param_opts_dflt = {
    0,
    0,
    NULL,
    NULL,
    0.0,
};

#define ST_OFF(f) M_ST_OFF(filter_param_opts_t,f)
static m_option_t filter_param_opts_fields[] = {
    {"id", ST_OFF(id), CONF_TYPE_INT, 0, 0 , 0, NULL},
    {"audio-video", ST_OFF(av), CONF_TYPE_INT, 0, 0 , 0, NULL},
    {"filter-name", ST_OFF(filtername), CONF_TYPE_STRING, 0 ,0 , 0, NULL},
    {"cmd", ST_OFF(cmd), CONF_TYPE_STRING, 0 ,0 , 0, NULL},
    {"pts", ST_OFF(pts), CONF_TYPE_FLOAT, 0 ,0 , 0, NULL},
    { NULL, NULL, 0, 0, 0, 0,  NULL }
};

static m_struct_t filter_param_opts = {
    "filter-cmd",
    sizeof(filter_param_opts_t),
    &filter_param_opts_dflt,
    filter_param_opts_fields
};

static struct filter_cmd_type_s {
    const char *name;
    int  cmdid;
} const filter_cmd_type_dflt[] = {
    {"dynamic",FILTER_CMD_START},
    {"static", FILTER_CMD_STOP},
    {NULL,0}
};

typedef struct dynamic_filter_cmd_s {
    int  insertid;
    int         av;
    const char *filter;
    filter_cmd_t *cmd;
} dynamic_filter_cmd_t;


dynamic_filter_cmd_t *dynamic_filter_cmd = NULL;

static int m_struct_set_options(m_struct_t* st, void* obj, char **param) {
    char *p;
    int r,n=0;

    while (*param) {
        if (!st->fields[n].name) return 0;
        if ((p=strchr(*param,':'))) *p++=0;
        if (strlen(*param)>0) {
                r = m_struct_set(st, obj, (char*) st->fields[n].name, *param);
            if (r<0) return r;
        }
        *param=p;
        n++;
    }
    return 0;
}

static dynamic_filter_cmd_t* seek_root_cmd(int id, int av, const char *filter) {
    int i;

    if(!dynamic_filter_cmd) {
        dynamic_filter_cmd=malloc(2*sizeof(dynamic_filter_cmd_t));
        dynamic_filter_cmd[0].insertid=id;
        dynamic_filter_cmd[0].av=av;
        dynamic_filter_cmd[0].filter=filter;
        dynamic_filter_cmd[0].cmd=NULL;
        dynamic_filter_cmd[1].insertid=0,
        dynamic_filter_cmd[1].av=av;
        dynamic_filter_cmd[1].filter=NULL;
        dynamic_filter_cmd[1].cmd=NULL;
        return &dynamic_filter_cmd[0];
    }
    for(i=0;;i++){
        if(!dynamic_filter_cmd[i].filter) break;
        if(dynamic_filter_cmd[i].insertid==id &&
           dynamic_filter_cmd[i].av==av &&
           dynamic_filter_cmd[i].filter==filter)
            return &dynamic_filter_cmd[i];
    }
    dynamic_filter_cmd=
        realloc(dynamic_filter_cmd,sizeof(dynamic_filter_cmd_t)*(i+2));
    dynamic_filter_cmd[i].insertid=id;
    dynamic_filter_cmd[i].av=av;
    dynamic_filter_cmd[i].filter=filter;
    dynamic_filter_cmd[i].cmd=NULL;
    dynamic_filter_cmd[i+1].insertid=0,
    dynamic_filter_cmd[i+1].av=av;
    dynamic_filter_cmd[i+1].filter=NULL;
    dynamic_filter_cmd[i+1].cmd=NULL;
    return &dynamic_filter_cmd[i];
}

static void *seek_insert_opts(dynamic_filter_cmd_t *root_cmd,
                              int type,
                              double pts) {
    filter_cmd_t *cmd;

    if (!root_cmd) return NULL;
    cmd=root_cmd->cmd;
    if (!cmd) return NULL;
    while(cmd) {
            if (cmd->type==type && cmd->pts==pts)
            return cmd->opts;
        cmd=cmd->next;
    }
    return NULL;
}

static void add_insert_cmd(dynamic_filter_cmd_t *root_cmd,
                           int type,
                           double pts,
                           void *opts) {
    filter_cmd_t *cmd, *prev=NULL;

    if (!root_cmd->cmd) {
        root_cmd->cmd=malloc(sizeof(filter_cmd_t));
        root_cmd->cmd->type=type;
        root_cmd->cmd->pts=pts;
        root_cmd->cmd->opts=opts;
        root_cmd->cmd->next=NULL;
        return;
    }
    cmd=root_cmd->cmd;
    while(cmd) {
            if (cmd->type==type && cmd->pts==pts)
            return;
        if ((cmd->pts<pts || (cmd->pts==pts && cmd->type<type)) &&
            (!prev ||
            (cmd->pts>prev->pts ||
            (cmd->pts==prev->pts && cmd->type>prev->type))))
            prev=cmd;
        cmd=cmd->next;
    }
    cmd=malloc(sizeof(filter_cmd_t));
    cmd->type=type;
    cmd->pts=pts;
    cmd->opts=opts;
    cmd->next=NULL;
    if (!prev) {
        cmd->next=root_cmd->cmd;
        root_cmd->cmd=cmd;
        return;
    }
    cmd->next=prev->next;
    prev->next=cmd;
}


static void add_insert_filters(char *param, int lineno) {
    char *pcurr, *tmp;
    int i, cmd_type=0, new_opts=0;
    filter_param_opts_t *config = NULL;
    void *opts;
    dynamic_filter_cmd_t *root_cmd = NULL;

    config = m_struct_alloc(&filter_param_opts);
    m_struct_reset(&filter_param_opts, config, NULL);
    tmp=strdup(param);
    pcurr=tmp;
    if ((m_struct_set_options(&filter_param_opts, config, &pcurr)<0)) {
        free(config);
        free(tmp);
        return;
    }
    if (config->id<=0) {
        if (lineno)
            mp_msg(MSGT_VFILTER,MSGL_ERR,
                   "Parse error #%d line. Illegal insertid: %d.\n",
                   lineno,config->id);
        else
            mp_msg(MSGT_VFILTER,MSGL_ERR,
                   "Parse error. Illegal insertid %d.\n",
                   config->id);
        free(config);
        free(tmp);
        return;
    }
    if (!config->filtername) {
        if (lineno)
            mp_msg(MSGT_VFILTER,MSGL_ERR,
                   "Parse error #%d line. Filter name required.\n",
                   lineno);
        else
            mp_msg(MSGT_VFILTER,MSGL_ERR,
                   "Parse error. Filter name required.\n");
        free(config);
        free(tmp);
        return;
    }
    if (!config->cmd) {
        if (lineno)
            mp_msg(MSGT_VFILTER,MSGL_ERR,
                   "Parse error #%d line. Filter command required.\n",
                   lineno);
        else
            mp_msg(MSGT_VFILTER,MSGL_ERR,
                   "Parse error. Filter command required.\n");
        free(config);
        free(tmp);
        return;
    }
    for(i=0;;i++){
        if(!filter_cmd_type_dflt[i].name) {
        if (lineno)
            mp_msg(MSGT_VFILTER,MSGL_ERR,
                   "Invalid filter command: '%s'. Line: %d.\n",
                   config->cmd,lineno);
        else
            mp_msg(MSGT_VFILTER,MSGL_ERR,
                   "Invalid filter command: '%s'.\n",
                   config->cmd);
        free(config);
        free(tmp);
        return;
        }
        if(!strcmp(filter_cmd_type_dflt[i].name,config->cmd)) break;
    }
    cmd_type=filter_cmd_type_dflt[i].cmdid;
    switch (config->av) {
    case FILTER_CMD_TYPE_VIDEO:
        for(i=0;;i++){
            if(!cmd_video_filter_list[i]) {
                mp_msg(MSGT_VFILTER,MSGL_ERR,
                       MSGTR_CouldNotFindVideoFilter,
                       config->filtername);
                free(config);
                free(tmp);
                return;
            }
            if(!strcmp(cmd_video_filter_list[i]->name,config->filtername))
                break;
        }
        root_cmd = seek_root_cmd(config->id,config->av,
                                 cmd_video_filter_list[i]->name);
        opts=seek_insert_opts(root_cmd,cmd_type,config->pts);
        if (!opts) {
            opts=m_struct_alloc((m_struct_t*)cmd_video_filter_list[i]->opts);
                m_struct_reset((m_struct_t*)cmd_video_filter_list[i]->opts,
                               opts, NULL);
            new_opts=1;
        }
        if ((m_struct_set_options((m_struct_t*)cmd_video_filter_list[i]->opts,
                                  opts, &pcurr)<0)) {
            free(opts);
            free(config);
            free(tmp);
            return;
        }
        free(tmp);
        break;
    case FILTER_CMD_TYPE_AUDIO:
        for(i=0;;i++){
            if(!cmd_audio_filter_list[i]) {
                mp_msg(MSGT_VFILTER,MSGL_ERR,
                       "Couldn't find audio filter '%s'.\n",
                       config->filtername);
                free(config);
                free(tmp);
                return;
            }
            if(!strcmp(cmd_audio_filter_list[i]->name,config->filtername))
                break;
        }
        root_cmd = seek_root_cmd(config->id,config->av,
                                 cmd_audio_filter_list[i]->name);
        opts=seek_insert_opts(root_cmd,cmd_type,config->pts);
        if (!opts) {
            opts=m_struct_alloc((m_struct_t*)cmd_audio_filter_list[i]->opts);
                m_struct_reset((m_struct_t*)cmd_audio_filter_list[i]->opts,
                               opts, NULL);
            new_opts=1;
        }
        if ((m_struct_set_options((m_struct_t*)cmd_audio_filter_list[i]->opts,
                                  opts, &pcurr)<0)) {
            free(opts);
            free(config);
            free(tmp);
            return;
        }
        free(tmp);
        break;
    default:
        if (lineno)
            mp_msg(MSGT_VFILTER,MSGL_ERR,
                   "Parse error #%d line. Illegal type: %d.\n",
                   lineno,config->av);
        else
            mp_msg(MSGT_VFILTER,MSGL_ERR,
                   "Parse error. Illegal type: %d.\n",
                   config->av);
        free(config);
        free(tmp);
        return;
    }
    if (new_opts)
        add_insert_cmd(root_cmd,cmd_type,config->pts,opts);
    free(config);
}

static void parse_dynamic_filters(char *param) {
    FILE *fh;
    char *buffer, *line, *p;
    int lineno=0;

    if((strncasecmp(param,"file:",5)))
        return add_insert_filters(param,0);
    param+=5;
    if (!(*param)) return;
    if (!(fh=fopen(param,"rb"))) {
        mp_msg(MSGT_VFILTER,MSGL_ERR,
               "Can't open insert filters file: %s.\n",
               param);
        return;
    }
    buffer=malloc(PARSE_BUFFER_SIZE+1);
    while (!feof(fh)) {
        memset(buffer,0,PARSE_BUFFER_SIZE);
        fgets(buffer,PARSE_BUFFER_SIZE,fh);
        lineno++;
        line=buffer;
        while (isspace(*line)) line++;
        if(*line==0 || *line=='#') continue;
        for(p=line+strlen(line)-1;isspace(*p)||*p=='\r'||*p=='\n';p--) *p=0;
        add_insert_filters(line,lineno);
    }
    free(buffer);
    fclose(fh);
}

int dynamicfilter_parseoption(const m_option_t* conf,
                       const char *opt,
                       const char *param){
    char *tmp, *ps, *pe;

    tmp=strdup(param);
    ps=tmp;
    while(ps) {
        if((pe=strchr(ps,','))) *pe++=0;
            parse_dynamic_filters(ps);
        ps=pe;
    }
    free(tmp);
    return 1;
}


filter_cmd_t* get_filter_cmd(int id, int av, const char *filter) {
    int i;

    if (!dynamic_filter_cmd) return NULL;
    for(i=0;;i++){
        if(!dynamic_filter_cmd[i].filter) break;
        if(dynamic_filter_cmd[i].insertid==id &&
           dynamic_filter_cmd[i].av==av &&
           dynamic_filter_cmd[i].filter==filter)
            return dynamic_filter_cmd[i].cmd;
    }
    return NULL;
}
