/*
 *  Import audio from shmem stream.
 *
 *  Based on af_pan.c and af_format.c
 *
 *  Copyright (C) 2007 Attila Ötvös <attila@onebithq.com>
 *
 *  MPlayer is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  MPlayer is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with MPlayer; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */
#include <stdio.h>
#include <stdlib.h>

#include <inttypes.h>
#include <math.h>
#include <limits.h>
#include <string.h>

// Integer to float conversion through lrintf()
#ifdef HAVE_LRINTF
#include <math.h>
long int lrintf(float);
#else
#define lrintf(x) ((int)(x))
#endif

#include "config.h"
#include "af.h"
#include "shmem_reader.h"
#include "m_struct.h"
#include "m_option.h"
#include "dynamicfilter.h"

#define FLOAT_NOSET        (int)(1<<31)

#define CMD_LEVEL0        0
#define CMD_LEVEL1        1

// Data for specific instances of this filter
typedef struct af_import_s
{
    int id;                 // shmem stream audio id
    float level0;                // Gain level for master channels
    float level1;                // Gain level for import channels
    void* audio_reader;         // import audio reader priv
    int err;                // Don't import if error
    long int samples;        // All filtered samples
    filter_cmd_t *cmd;
    filter_cmd_t *s_level0, *e_level0;
    filter_cmd_t *s_level1, *e_level1;
} af_import_t;


typedef struct afcmd_priv_s {
    float level0;
    float level1;
} afcmd_priv_t;

static afcmd_priv_t const afcmd_priv_dflt = {
    FLOAT_NOSET,                                  // level0
    FLOAT_NOSET                                   // level1
};

static float get_cmd_var(afcmd_priv_t *cmd_priv, int varid) {
    switch (varid) {
    case CMD_LEVEL0:
        return cmd_priv->level0;
    case CMD_LEVEL1:
        return cmd_priv->level1;
    }
    return FLOAT_NOSET;
}

static void init_dynvar(struct af_instance_s* af,
                        filter_cmd_t **start,
                        filter_cmd_t **end,
                        int varid) {
    af_import_t* s = af->setup;
    filter_cmd_t *cmd = s->cmd;
    float var;

    while (cmd) {
        var=get_cmd_var(cmd->opts,varid);
        if ((*start) && !(*end) && var!=FLOAT_NOSET) {
            *end=cmd;
            return;
        }
        if (!(*start) && var!=FLOAT_NOSET) *start=cmd;
        cmd=cmd->next;
    }
}

static void init_dynvars(struct af_instance_s* af) {
    af_import_t* s = af->setup;
    init_dynvar(af,&s->s_level0,&s->e_level0, CMD_LEVEL0);
    init_dynvar(af,&s->s_level1,&s->e_level1, CMD_LEVEL1);
}

static void cmd_to_dynvar(double pts, filter_cmd_t **start,
                          filter_cmd_t **end, int varid) {
    filter_cmd_t *cmd = *end;
    float var;

    if (!cmd) return;
    if (cmd->pts>=pts) return;
    *start=*end;
    cmd=(*end)->next;
    while (cmd) {
        var=get_cmd_var(cmd->opts,varid);
        if (var!=FLOAT_NOSET) {
            *end=cmd;
            return;
        }
        cmd=cmd->next;
    }
    *end=NULL;
}

static void cmd_to_dynvars(struct af_instance_s* af, double pts) {
    af_import_t* s = af->setup;
    cmd_to_dynvar(pts,&s->s_level0,&s->e_level0,CMD_LEVEL0);
    cmd_to_dynvar(pts,&s->s_level1,&s->e_level1,CMD_LEVEL1);
}

static int dynvar_to_param(double pts, float *var, filter_cmd_t **start,
                           filter_cmd_t **end, int varid) {
    double rate,ds,de,dr;
    float new, ret = 0;

    while (1) {
        if (!(*start)) return 0;
        if ((*start)->pts>pts) return 0;
        if (!(*end) || ((*start)->pts==(*end)->pts &&
            (*start)->type!=FILTER_CMD_STOP)) {
            if (*var==get_cmd_var((*start)->opts,varid))
                return 0;
            *var=get_cmd_var((*start)->opts,varid);
            return 1;
        }
        if ((*end)->type==FILTER_CMD_STOP) {
            if (*var!=get_cmd_var((*start)->opts,varid)) {
                *var=get_cmd_var((*start)->opts,varid);
                ret=1;
            }
            if ((*start)->pts!=(*end)->pts)
                return ret;
            *start=*end;
            *end=(*end)->next;
            continue;
        }
        break;
    }
    rate=(pts-(*start)->pts)/((*end)->pts-(*start)->pts);
    if (rate>1.0) rate=1.0;
    ds=get_cmd_var((*start)->opts,varid);
    de=get_cmd_var((*end)->opts,varid);
    dr=(de-ds)*rate+ds;
    new=dr;
    if (*var==new) return ret;
    *var=new;
    return 1;
}

static int dynvars_to_param(struct af_instance_s* af, double pts) {
    af_import_t* s = af->setup;
    int changed = 0;
    changed|=dynvar_to_param(pts,&s->level0,&s->s_level0,
                             &s->e_level0,CMD_LEVEL0);
    changed|=dynvar_to_param(pts,&s->level1,&s->s_level0,
                             &s->e_level1,CMD_LEVEL1);
    return changed;
}



// Initialization and runtime control
static int control(struct af_instance_s* af, int cmd, void* arg)
{
    af_import_t* s = af->setup;

    switch(cmd){
    case AF_CONTROL_REINIT:
        // Sanity check
        if(!arg) return AF_ERROR;
        af->data->rate        = ((af_data_t*)arg)->rate;
        af->data->nch        = ((af_data_t*)arg)->nch;
        af->data->format= AF_FORMAT_FLOAT_NE;
        af->data->bps        = 4;
        return af_test_output(af,(af_data_t*)arg);
    case AF_CONTROL_COMMAND_LINE:{
        s->id=0;
        s->level0=1.0;
        s->level1=1.0;
        s->audio_reader=NULL;
        sscanf((char*)arg,"%i:%f:%f", &s->id,&s->level0,&s->level1);
        s->level0=clamp(s->level0,0.0f,1.0f);
        s->level1=clamp(s->level1,0.0f,1.0f);
        s->audio_reader=open_audio_shmemreader(s->id);
        s->cmd=get_filter_cmd(s->id,
                              FILTER_CMD_TYPE_AUDIO,
                              af->info->name);
        init_dynvars(af);
        return AF_OK;
        }
    }
    return AF_UNKNOWN;
}

// Deallocate memory
static void uninit(struct af_instance_s* af)
{
    af_import_t* s = af->setup;

    if(af->data)
        free(af->data->audio);
    free(af->data);
    if(af->setup) {
        if (s->audio_reader)
            close_audio_shmemreader(s->audio_reader);
        free(af->setup);
    }
}

static inline uint32_t load24bit(void* data, int pos) {
#if WORDS_BIGENDIAN
    return (((uint32_t)((uint8_t*)data)[3*pos])<<24) |
           (((uint32_t)((uint8_t*)data)[3*pos+1])<<16) |
           (((uint32_t)((uint8_t*)data)[3*pos+2])<<8);
#else
    return (((uint32_t)((uint8_t*)data)[3*pos])<<8) |
           (((uint32_t)((uint8_t*)data)[3*pos+1])<<16) |
           (((uint32_t)((uint8_t*)data)[3*pos+2])<<24);
#endif
}

static inline void store24bit(void* data, int pos, uint32_t expanded_value) {
#if WORDS_BIGENDIAN
    ((uint8_t*)data)[3*pos]=expanded_value>>24;
    ((uint8_t*)data)[3*pos+1]=expanded_value>>16;
    ((uint8_t*)data)[3*pos+2]=expanded_value>>8;
#else
    ((uint8_t*)data)[3*pos]=expanded_value>>8;
    ((uint8_t*)data)[3*pos+1]=expanded_value>>16;
    ((uint8_t*)data)[3*pos+2]=expanded_value>>24;
#endif
}

static void float2int(float* in, void* out, int len, int bps)
{
    register int i;

    switch(bps){
    case(1):
        for(i=0;i<len;i++)
            ((int8_t*)out)[i] = lrintf(127.0 * in[i]);
        break;
    case(2):
        for(i=0;i<len;i++)
            ((int16_t*)out)[i] = lrintf(32767.0 * in[i]);
        break;
    case(3):
        for(i=0;i<len;i++)
            store24bit(out, i, lrintf(2147483647.0 * in[i]));
        break;
    case(4):
        for(i=0;i<len;i++)
          ((int32_t*)out)[i] = lrintf(2147483647.0 * in[i]);
        break;
    }
}

static void int2float(void* in, float* out, int len, int bps)
{
    register int i;

    switch(bps){
    case(1):
        for(i=0;i<len;i++)
            out[i]=(1.0/128.0)*((int8_t*)in)[i];
        break;
    case(2):
        for(i=0;i<len;i++)
            out[i]=(1.0/32768.0)*((int16_t*)in)[i];
        break;
    case(3):
        for(i=0;i<len;i++)
            out[i]=(1.0/2147483648.0)*((int32_t)load24bit(in, i));
        break;
    case(4):
        for(i=0;i<len;i++)
            out[i]=(1.0/2147483648.0)*((int32_t*)in)[i];
        break;
    }
}


// Filter data through filter
static af_data_t* play(struct af_instance_s* af, af_data_t* data)
{
    af_data_t*    c0   = data;            // Current working data
    af_data_t*    c1   = NULL;            // Import data
    af_data_t*    cl   = NULL;            // Import data (converted)
    af_import_t*  s    = af->setup;       // Setup for this instance
    float*        inf0 = c0->audio;       // Input audio data (float)
    int16_t*      ini0 = c0->audio;       // Input audio data (int)
    float*        inf1 = NULL;            // Import audio data (float)
    int16_t*      ini1 = c0->audio;       // Import audio data (int)
    float*        endf0 = inf0+c0->len/4; // End of loop (int)
    int16_t*      endi0 = ini0+c0->len/2; // End of loop (float)
    float*        endf1 = NULL;           // End of loop (float)
    int16_t*      endi1 = NULL;           // End of loop (int)
    register int  j,l=0;
    int           samples = 0;
    double        pts = 0.0;

    if(c0->format==AF_FORMAT_S16_NE) {
        samples=c0->len/(sizeof(int16_t)*c0->nch);
    }
    if(c0->format==AF_FORMAT_FLOAT_NE) {
        samples=c0->len/(sizeof(float)*c0->nch);
    }
    if(s->err) {
        s->samples+=samples;
        return c0;
    }
    pts=(float)s->samples/(float)c0->rate;
    c1=(af_data_t*)get_shmemreader_audio_data(s->audio_reader,samples,pts);
    if(!c1) {
        s->samples+=samples;
        return c0;
    }
    cmd_to_dynvars(af,pts);
    dynvars_to_param(af,pts);
    if(c0->rate!=c1->rate) {
        printf("#%d. Can't import audio because sample rate is bad: %d"
               " Hz (need: %d Hz)\n",
               s->id,c1->rate,c0->rate);
        s->err=1;
        s->samples+=samples;
        return c0;
    }
    if(c0->format==AF_FORMAT_S16_NE && c1->format==AF_FORMAT_S16_NE)
        cl=c1;
    if(c0->format==AF_FORMAT_FLOAT_NE && c1->format==AF_FORMAT_S16_NE) {
        cl=malloc(sizeof(af_data_t));
        cl->len=c1->len*2;
        cl->audio=malloc(cl->len);
        cl->rate=c1->rate;
        cl->nch=c1->nch;
        cl->format=AF_FORMAT_FLOAT_NE;
        cl->bps=4;
        int2float(c1->audio, cl->audio, c1->len/2, 2);
    }
    if(c0->format==AF_FORMAT_S16_NE && c1->format==AF_FORMAT_FLOAT_NE) {
        cl=malloc(sizeof(af_data_t));
        cl->len=c1->len/2;
        cl->audio=malloc(cl->len);
        cl->rate=c1->rate;
        cl->nch=c1->nch;
        cl->format=AF_FORMAT_S16_NE;
        cl->bps=2;
        float2int(c1->audio, cl->audio, c1->len/2, 2);
    }
    if(c0->format==AF_FORMAT_FLOAT_NE && c1->format==AF_FORMAT_FLOAT_NE)
        cl=c1;
    if(!cl) {
        s->samples+=samples;
        return c0;
    }
    if(cl->format==AF_FORMAT_S16_NE) {
        ini1=cl->audio;
        endi1=ini1+cl->len/2;
    }
    if(cl->format==AF_FORMAT_FLOAT_NE) {
        inf1=cl->audio;
        endf1=inf1+cl->len/4;
    }

    if(c0->format==AF_FORMAT_S16_NE) {
        while(ini0 < endi0){
            if(ini1>=endi1) break;
            for(j=0;j<min(c0->nch,c1->nch);j++){
                register float  x   = 0.0;
                x = ini0[j] * s->level0 + ini1[j] * s->level1;
                inf0[j] = x;
            }
        ini0+= c0->nch;
        ini1+= c1->nch;
        l++;
        }
    }
    if(c0->format==AF_FORMAT_FLOAT_NE) {
        while(inf0 < endf0){
            if(inf1>=endf1) break;
            for(j=0;j<min(c0->nch,c1->nch);j++){
            register float  x   = 0.0;
            x = inf0[j] * s->level0 + inf1[j] * s->level1;
            inf0[j] = x;
        }
        inf0+= c0->nch;
        inf1+= c1->nch;
        l++;
        }
    }
    if(c1->format==AF_FORMAT_S16_NE)
        drop_shmemreader_audio_data(s->audio_reader,c1,
                                    sizeof(int16_t)*l*c1->nch);
    if(c1->format==AF_FORMAT_FLOAT_NE)
        drop_shmemreader_audio_data(s->audio_reader,c1,
                                    sizeof(float)*l*c1->nch);
    if(c1!=cl) {
        free(cl->audio);
        free(cl);
    }
    s->samples+=samples;
    return c0;
}

// Allocate memory and set function pointers
static int af_open(af_instance_t* af){
    af->control=control;
    af->uninit=uninit;
    af->play=play;
    af->mul=1;
    af->data=calloc(1,sizeof(af_data_t));
    af->setup=calloc(1,sizeof(af_import_t));
    if(af->data == NULL || af->setup == NULL)
        return AF_ERROR;
    return AF_OK;
}

#undef ST_OFF
#define ST_OFF(f) M_ST_OFF(struct afcmd_priv_s,f)
static m_option_t afcmd_opts_fields[] = {
    {"level-master", ST_OFF(level0), CONF_TYPE_FLOAT, 0 ,0 , 0, NULL},
    {"level-import", ST_OFF(level1), CONF_TYPE_FLOAT, 0 ,0 , 0, NULL},
    { NULL, NULL, 0, 0, 0, 0,  NULL }
};

static m_struct_t afcmd_opts = {
    "import",
    sizeof(struct afcmd_priv_s),
    &afcmd_priv_dflt,
    afcmd_opts_fields
};


// Description of this filter command
dfcmd_info_t afcmd_info_import = {
    "Audio import filter",
    "import",
    "Otvos Attila",
    "",
    &afcmd_opts
};


// Description of this filter
af_info_t af_info_import = {
    "Import audio filter",
    "import",
    "Otvos Attila",
    "",
    AF_FLAGS_REENTRANT,
    af_open
};
