/*
 *  Mixer two mp_image.
 *
 *  Copyright (C) 2007 Attila Ötvös <attila@onebithq.com>
 *
 *  MPlayer is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  MPlayer is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with MPlayer; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#include <string.h>

#include "config.h"
#include "img_format.h"
#include "mp_image.h"
#include "libavutil/common.h"
#include "libvo/fastmemcpy.h"

/**
 * \brief imgsize() returns the mpi planes size
 * \param mpi: mp_image_t pointer
 * \return full planes size
 */
int imgsize(mp_image_t *mpi) {
    return mpi->width*mpi->height*mpi->bpp/8;
}

/**
 * \brief imgscalex() returns the mpi planes[0] bytes per pixel
 * \param mpi: mp_image_t pointer
 * \return planes[0] bytes per pixel
 */
static int imgscalex(mp_image_t *mpi) {

    switch(mpi->imgfmt){
    case IMGFMT_YUY2:
    case IMGFMT_UYVY:
        return 2;
    case IMGFMT_BGR24:
    case IMGFMT_RGB24:
        return 3;
    }
    return 1;
}

/**
 * \brief cpy_pic_alpha() insert line from src image to dst image
 * \param dst: destination image line
 * \param src: source image line
 * \param len: line width
 * \param alpha: src alpha (0 - opacity 255 - )
 */
static void cpy_pic_alpha(unsigned char *dst,
                          unsigned char *src,
                          int len,
                          int alpha) {
    int i;
    if(!alpha) return;
    if(alpha>=255 || alpha<0) {
        fast_memcpy(dst,src,len);
        return;
    }
    for(i=0;i<len;i++) dst[i]=(dst[i]*(255-alpha)+src[i]*alpha)/255;
}

/**
 * \brief cut_image() cut part image from src to dst
 * \param src: source image
 * \param dst: destination image
 * \param x,y: cut top right corner
 * \param w,h: cut image part width and height
 */
void cut_image(mp_image_t* src,mp_image_t* dst, int x, int y, int w, int h) {
    int posy;

    for(posy=0;posy<h;posy++) {
        fast_memcpy( dst->planes[0] +
                    (posy*dst->stride[0]),
                    src->planes[0] + ((posy+y)*src->stride[0]) + x,
                    w * imgscalex(src));
        if (posy%2 && src->chroma_height) {
            fast_memcpy( dst->planes[1] + ((posy/2)*dst->stride[1]),
                        src->planes[1] + ((((posy+y)/2)*(src->w)/2)) + x/2,
                        w/2);
            fast_memcpy( dst->planes[2] + ((posy/2)*dst->stride[2]),
                        src->planes[2] + ((((posy+y)/2)*(src->w)/2)) + x/2,
                        w/2);
        }
        if (src->chroma_height==0) {
            if (src->planes[1])
                fast_memcpy( dst->planes[1] + (posy*dst->stride[1]),
                                src->planes[1] + ((posy+y)*src->w) + x,
                                w);
            if (src->planes[2])
                fast_memcpy( dst->planes[2] + (posy*dst->stride[2]),
                                src->planes[2] + ((posy+y)*src->w) + x,
                                w);
        }
    }


}

/**
 * \brief mp_image_mixer() mixing mpi image to dmpi image
 * \param dmpi   : destination image
 * \param mpi    : inserted image
 * \param window : target window on dmpi (x,y: positon, w,h: size)
 * \param scaled : mpi size (w,h) and position (x,y)
 * \param alpha  : insert image alpha
 */
void mp_image_mixer(mp_image_t* dmpi,
                    mp_image_t *mpi,
                    int window_x, int window_y, int window_w, int window_h,
                    int scaled_x, int scaled_y, int scaled_w, int scaled_h,
                    int alpha) {

    int line, dst_line, src_line;
    int width, height;
    int chroma_width, chroma_height;
    int dst_x, dst_y;
    int src_x, src_y;

    if(!alpha) return;

// dest
    dst_x=FFMAX(window_x,scaled_x);
    dst_y=FFMAX(window_y,scaled_y);
    width=FFMIN(window_w,dmpi->w-dst_x);
    height=FFMIN(window_h,dmpi->h-dst_y);
// source
    src_x=FFMAX(0,-scaled_x);
    src_y=FFMAX(0,-scaled_y);
    width=FFMIN(width,mpi->w-src_x);
    height=FFMIN(height,mpi->h-src_y);
// chroma
    width=FFMIN(width,scaled_w-src_x);
    height=FFMIN(height,scaled_h-src_y);
    chroma_width=(mpi->chroma_width * width / mpi->w) & ~1;
    chroma_height=(mpi->chroma_height * height / mpi->h) & ~1;

    if(width<0 || height<0) return;
    if (!dst_x && !dst_y && !src_x && !src_y && (alpha>=255 || alpha<0)) {
        memcpy_pic(dmpi->planes[0], mpi->planes[0], width * imgscalex(mpi),
                   height, dmpi->stride[0], mpi->stride[0]);
        if (mpi->planes[1])
            memcpy_pic(dmpi->planes[1], mpi->planes[1], chroma_width,
                       chroma_height, dmpi->stride[1], mpi->stride[1]);
        if (mpi->planes[2])
            memcpy_pic(dmpi->planes[2], mpi->planes[2], chroma_width,
                       chroma_height, dmpi->stride[2], mpi->stride[2]);
    } else {
    for( line = 0 ; line < height ; line++ ) {
        dst_line = dst_y + line;
        src_line = src_y + line;
        cpy_pic_alpha(dmpi->planes[0] +
                      (dst_line*dmpi->stride[0]) + dst_x * imgscalex(mpi),
                      mpi->planes[0] + (src_line*mpi->stride[0]) + src_x,
                      width * imgscalex(mpi),
                      alpha);
        if (line%2 && chroma_height) {
            cpy_pic_alpha(dmpi->planes[1] + ((dst_line>>1)*dmpi->stride[1]) +
                          (dst_x>>1),
                           mpi->planes[1] + ((src_line>>1)*mpi->stride[1]) +
                           (src_x>>1),
                           chroma_width,
                           alpha);
            cpy_pic_alpha(dmpi->planes[2] + ((dst_line>>1)*dmpi->stride[2]) +
                          (dst_x>>1),
                          mpi->planes[2] + ((src_line>>1)*mpi->stride[2]) +
                          (src_x>>1),
                          chroma_width,
                          alpha);
        }
        if (mpi->chroma_height==0) {
            if (mpi->planes[1])
                cpy_pic_alpha(dmpi->planes[1] + (dst_line*dmpi->stride[1]) +
                              dst_x,
                              mpi->planes[1] + (src_line*mpi->w) + src_x,
                              width,
                              alpha);
            if (mpi->planes[2])
                cpy_pic_alpha(dmpi->planes[2] + (dst_line*dmpi->stride[2]) +
                              dst_x,
                              mpi->planes[2] + (src_line*mpi->w) + src_x,
                              width,
                              alpha);
            }
        }
    }
}
