/*
 *  Export video frame to shared memory.
 *
 *  Based on vf_screenshot.c and vf_expand.c
 *
 *
 *  Copyright (C) 2007 Attila Ötvös <attila@onebithq.com>
 *
 *  MPlayer is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  MPlayer is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with MPlayer; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#include <string.h>
#include <inttypes.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "mp_msg.h"

#include "img_format.h"
#include "mp_image.h"
#include "vf.h"
#include "vf_scale.h"

#include "m_option.h"
#include "m_struct.h"

#include "libswscale/swscale.h"
#include "stream/stream.h"
#include "stream/shmem_common.h"
#include "stream/stream_shmem.h"
#include "libmpdemux/demuxer.h"

#define OSD_SUPPORT

#define OSD_OFF     0
#define OSD_ON      1
#define OSD_SCALE   2

#ifdef OSD_SUPPORT
#include "libvo/sub.h"
#include "libvo/osd.h"

static struct vf_instance_s* vf=NULL; // fixme (needs sub.c changes)
static mp_image_t* osdmpi=NULL;
#endif

static struct vf_priv_s {
    int store_slices;
    int w, h;
    int size;
    mp_image_t* dst;
    mp_image_t* odst;
    struct SwsContext *ctx;
    struct SwsContext *ctx2;
    int shmem_opened;
    stream_t* stream;
    shmem_video_t *video;
    int shmemid;
    int fmt;
    int ow, oh, ofmt;
    int osd;
} const vf_priv_dflt = {
    0,
    0, 0,
    0,
    NULL,
    NULL,
    NULL,
    NULL,
    0,
    NULL,
    1000,
    0,
    0, 0, 0,
    0
};

//===========================================================================//

static int config(struct vf_instance_s* vf,
                  int width, int height, int d_width, int d_height,
                  unsigned int flags, unsigned int outfmt) {

    if (!vf->priv->fmt)
    switch(outfmt){
        case IMGFMT_YV12:
        case IMGFMT_I420:
        case IMGFMT_IYUV:
        case IMGFMT_UYVY:
        case IMGFMT_YUY2:
        case IMGFMT_BGR24:
        case IMGFMT_RGB24:
        case IMGFMT_Y800:
        case IMGFMT_Y8:
        vf->priv->fmt=outfmt;
        break;
    }
    if (!vf->priv->fmt) vf->priv->fmt=IMGFMT_BGR24;

    if (!vf->priv->w) vf->priv->w = d_width;
    if (!vf->priv->h) vf->priv->h = d_height;

    vf->priv->ow = d_width;
    vf->priv->oh = d_height;
    vf->priv->ofmt = outfmt;
#ifdef OSD_SUPPORT
    if (vf->priv->osd==OSD_SCALE)
        vf->priv->ctx2 = sws_getContextFromCmdLine(width, height, outfmt,
                                                   d_width, d_height, outfmt);
#endif
    vf->priv->ctx = sws_getContextFromCmdLine(width, height, outfmt,
                                              vf->priv->w, vf->priv->h,
                                              vf->priv->fmt);

    if (vf->priv->dst) free_mp_image(vf->priv->dst);
    if (vf->priv->odst) free_mp_image(vf->priv->odst);
    vf->priv->dst = NULL;
    vf->priv->odst = NULL;

    return vf_next_config(vf, width, height, d_width, d_height, flags, outfmt);
}

static int imgsize(mp_image_t *mpi) {

    switch(mpi->imgfmt){
    case IMGFMT_I420:
    case IMGFMT_IYUV:
    case IMGFMT_YV12:
        return mpi->w*mpi->h+2*(mpi->w>>1)*(mpi->h>>1);
    case IMGFMT_YUY2:
    case IMGFMT_UYVY:
        return mpi->w*mpi->h*2;
    case IMGFMT_Y800:
    case IMGFMT_Y8:
        return mpi->w*mpi->h;
    case IMGFMT_BGR24:
    case IMGFMT_RGB24:
        return mpi->w * mpi->h * ((IMGFMT_RGB_DEPTH(mpi->imgfmt) + 7) >> 3);
    }
    return 0;
}


static void write_shmem(struct vf_instance_s* vf, mp_image_t *mpi) {
    char fname[64];

    if (vf->priv->shmem_opened==0 && vf->priv->shmemid>0) {
        vf->priv->shmem_opened = 1;
        snprintf(fname,sizeof(fname),"shmem://%d",vf->priv->shmemid);
        set_shmem_size(SHMEM_TYPE_RAWDATA, imgsize(mpi)+sizeof(shmem_video_t));
        vf->priv->video=calloc(1,imgsize(mpi)+sizeof(shmem_video_t));
        vf->priv->video->id=SHMEM_VIDEO_PACKET;
        vf->priv->video->format=mpi->imgfmt;
        vf->priv->video->width=mpi->width;
        vf->priv->video->height=mpi->height;
        vf->priv->video->size=imgsize(mpi);
        vf->priv->stream = open_output_stream(fname, NULL);
    }
    if (vf->priv->stream) {
        memcpy(vf->priv->video->data,mpi->planes[0],imgsize(mpi));
        stream_write_buffer(vf->priv->stream, vf->priv->video,
                            imgsize(mpi)+sizeof(shmem_video_t));
    }
}

#ifdef OSD_SUPPORT
static void draw_func(int x0,int y0, int w,int h,unsigned char* src,
                        unsigned char *srca, int stride) {
    unsigned char* dst;

    if (w<=0 || h<=0) return; // nothing to do...

    dst = osdmpi->planes[0]+osdmpi->stride[0]*y0+(osdmpi->bpp>>3)*x0;

    switch(osdmpi->imgfmt){
    case IMGFMT_BGR15:
    case IMGFMT_RGB15:
        vo_draw_alpha_rgb15(w, h, src, srca, stride, dst, osdmpi->stride[0]);
        break;
    case IMGFMT_BGR16:
    case IMGFMT_RGB16:
        vo_draw_alpha_rgb16(w, h, src, srca, stride, dst, osdmpi->stride[0]);
        break;
    case IMGFMT_BGR24:
    case IMGFMT_RGB24:
        vo_draw_alpha_rgb24(w, h, src, srca, stride, dst, osdmpi->stride[0]);
        break;
    case IMGFMT_BGR32:
    case IMGFMT_RGB32:
        vo_draw_alpha_rgb32(w, h, src, srca, stride, dst, osdmpi->stride[0]);
        break;
    case IMGFMT_YV12:
    case IMGFMT_I420:
    case IMGFMT_IYUV:
    case IMGFMT_YVU9:
    case IMGFMT_IF09:
    case IMGFMT_Y800:
    case IMGFMT_Y8:
        vo_draw_alpha_yv12(w, h, src, srca, stride, dst, osdmpi->stride[0]);
        break;
    case IMGFMT_YUY2:
        vo_draw_alpha_yuy2(w, h, src, srca, stride, dst, osdmpi->stride[0]);
        break;
    case IMGFMT_UYVY:
        vo_draw_alpha_yuy2(w, h, src, srca, stride, dst+1, osdmpi->stride[0]);
        break;
    }
}

static void draw_osd(struct vf_instance_s* vf_,mp_image_t *mpi) {

    vf = vf_;
    osdmpi = mpi;
    vo_draw_text(mpi->w, mpi->h, draw_func);
}
#endif

static void scale_image(struct vf_instance_s* vf, mp_image_t *mpi) {

    if (!vf->priv->ctx) return;
    if (!vf->priv->dst) vf->priv->dst = alloc_mpi(vf->priv->w, vf->priv->h,
                                                  vf->priv->fmt);
#ifdef OSD_SUPPORT
    if (vf->priv->osd==OSD_SCALE) {
        if (!vf->priv->odst) vf->priv->odst = alloc_mpi(vf->priv->ow,
                                                        vf->priv->oh,
                                                        vf->priv->ofmt);
        sws_scale_ordered(vf->priv->ctx2, mpi->planes, mpi->stride, 0, mpi->h,
                          vf->priv->odst->planes, vf->priv->odst->stride);
        } else
#endif
        sws_scale_ordered(vf->priv->ctx, mpi->planes, mpi->stride, 0, mpi->h,
                          vf->priv->dst->planes, vf->priv->dst->stride);
}

static void start_slice(struct vf_instance_s* vf, mp_image_t *mpi) {
                        vf->dmpi=vf_get_image(vf->next,mpi->imgfmt,
                        mpi->type, mpi->flags, mpi->width, mpi->height);

    if (!vf->priv->dst) vf->priv->dst = alloc_mpi(vf->priv->w, vf->priv->h,
                                                  vf->priv->fmt);
#ifdef OSD_SUPPORT
    if (!vf->priv->odst && vf->priv->osd==OSD_SCALE)
        vf->priv->odst = alloc_mpi(vf->priv->ow, vf->priv->oh, vf->priv->ofmt);
#endif
    vf->priv->store_slices = 1;
}

static void draw_slice(struct vf_instance_s* vf,
        unsigned char** src, int* stride, int w,int h, int x, int y) {

    if (vf->priv->store_slices && vf->priv->ctx) {
        if (!vf->priv->dst) vf->priv->dst = alloc_mpi(vf->priv->w, vf->priv->h,
                                                      vf->priv->fmt);
#ifdef OSD_SUPPORT
        if (vf->priv->osd==OSD_SCALE) {
            if (!vf->priv->odst) vf->priv->odst = alloc_mpi(vf->priv->ow,
                                                            vf->priv->oh,
                                                            vf->priv->ofmt);
            sws_scale_ordered(vf->priv->ctx2, src, stride, y, h,
                              vf->priv->odst->planes,
                              vf->priv->odst->stride);
            } else
#endif
            sws_scale_ordered(vf->priv->ctx, src, stride, y, h,
                              vf->priv->dst->planes,
                              vf->priv->dst->stride);
        }
    vf_next_draw_slice(vf, src, stride, w, h, x, y);
}

static void get_image(struct vf_instance_s* vf, mp_image_t *mpi) {

    // FIXME: should vf.c really call get_image when using slices??
    if (mpi->flags & MP_IMGFLAG_DRAW_CALLBACK)
        return;
    vf->dmpi= vf_get_image(vf->next, mpi->imgfmt,
                           mpi->type, mpi->flags/* | MP_IMGFLAG_READABLE*/,
                           mpi->width, mpi->height);

    mpi->planes[0] = vf->dmpi->planes[0];
    mpi->stride[0] = vf->dmpi->stride[0];
    if (mpi->flags&MP_IMGFLAG_PLANAR){
        mpi->planes[1] = vf->dmpi->planes[1];
        mpi->planes[2] = vf->dmpi->planes[2];
        mpi->stride[1] = vf->dmpi->stride[1];
        mpi->stride[2] = vf->dmpi->stride[2];
    }
    mpi->width = vf->dmpi->width;
    mpi->flags |= MP_IMGFLAG_DIRECT;
    mpi->priv = (void*)vf->dmpi;
}

static int put_image(struct vf_instance_s* vf, mp_image_t *mpi, double pts) {
    mp_image_t *dmpi = (mp_image_t *)mpi->priv;

    if (mpi->flags & MP_IMGFLAG_DRAW_CALLBACK)
        dmpi = vf->dmpi;
    else if (!(mpi->flags&MP_IMGFLAG_DIRECT)){
        dmpi = vf_get_image(vf->next,mpi->imgfmt,
                            MP_IMGTYPE_EXPORT, 0,
                            mpi->width, mpi->height);
        vf_clone_mpi_attributes(dmpi, mpi);
        dmpi->planes[0] = mpi->planes[0];
        dmpi->planes[1] = mpi->planes[1];
        dmpi->planes[2] = mpi->planes[2];
        dmpi->stride[0] = mpi->stride[0];
        dmpi->stride[1] = mpi->stride[1];
        dmpi->stride[2] = mpi->stride[2];
        dmpi->width = mpi->width;
        dmpi->height = mpi->height;
    }

    if (vf->priv->ctx) {
        if (!vf->priv->store_slices) scale_image(vf, dmpi);
#ifdef OSD_SUPPORT
        switch(vf->priv->osd) {
        case OSD_ON:
            draw_osd(vf, vf->priv->dst);
            break;
        case OSD_SCALE:
            draw_osd(vf, vf->priv->odst);
            sws_scale_ordered(vf->priv->ctx, vf->priv->odst->planes,
                              vf->priv->odst->stride, 0,
                              vf->priv->odst->h, vf->priv->dst->planes,
                              vf->priv->dst->stride);
            break;
        }
#endif
        write_shmem(vf, vf->priv->dst);
        vf->priv->store_slices = 0;
    }
    return vf_next_put_image(vf, dmpi, pts);
}

//===========================================================================//
static int control(struct vf_instance_s* vf, int request, void* data) {

#ifdef OSD_SUPPORT
    switch(request){
    case VFCTRL_DRAW_OSD:
    if (vf->priv->osd) return CONTROL_TRUE;
    }
#endif
    return vf_next_control(vf, request, data);
}

static int query_format(struct vf_instance_s* vf, unsigned int fmt) {

    switch(fmt){
    case IMGFMT_YV12:
    case IMGFMT_I420:
    case IMGFMT_IYUV:
    case IMGFMT_UYVY:
    case IMGFMT_YUY2:
    case IMGFMT_BGR24:
    case IMGFMT_RGB24:
    case IMGFMT_Y800:
    case IMGFMT_Y8:
    case IMGFMT_BGR32:
    case IMGFMT_BGR16:
    case IMGFMT_BGR15:
    case IMGFMT_RGB32:
    case IMGFMT_YVU9:
    case IMGFMT_IF09:
    case IMGFMT_444P:
    case IMGFMT_422P:
    case IMGFMT_411P:
        return vf_next_query_format(vf, fmt);
    }
    return 0;
}

static void uninit(vf_instance_t *vf);
// open conflicts with stdio.h at least under MinGW

static int export_open(vf_instance_t *vf, char* args) {

    vf->config = config;
    vf->put_image = put_image;
    vf->query_format = query_format;
    vf->start_slice = start_slice;
    vf->draw_slice = draw_slice;
    vf->get_image = get_image;
    vf->uninit = uninit;
    vf->priv->w = (vf->priv->w+1)&~1;
    vf->priv->h = (vf->priv->h+1)&~1;
    return 1;
}

static void uninit(vf_instance_t *vf) {

    if (vf->priv->stream) free_stream(vf->priv->stream);
    if (vf->priv->ctx) sws_freeContext(vf->priv->ctx);
    if (vf->priv->ctx2) sws_freeContext(vf->priv->ctx2);
    if (vf->priv->dst) free_mp_image(vf->priv->dst);
    if (vf->priv->odst) free_mp_image(vf->priv->odst);
    if (vf->priv->video) free(vf->priv->video);
    free(vf->priv);
}

#define ST_OFF(f) M_ST_OFF(struct vf_priv_s,f)
static m_option_t vf_opts_fields[] = {
    {"id", ST_OFF(shmemid), CONF_TYPE_INT, 0 , 0, 0, NULL},
    {"fmt", ST_OFF(fmt), CONF_TYPE_IMGFMT, 0,0 ,0, NULL},
    {"w", ST_OFF(w), CONF_TYPE_INT, M_OPT_MIN,16,0, NULL},
    {"h", ST_OFF(h), CONF_TYPE_INT, M_OPT_MIN,16,0, NULL},
    {"osd", ST_OFF(osd), CONF_TYPE_INT, CONF_RANGE , 0, 2, NULL},
    { NULL, NULL, 0, 0, 0, 0,  NULL }
};

static m_struct_t vf_opts = {
    "export",
    sizeof(struct vf_priv_s),
    &vf_priv_dflt,
    vf_opts_fields
};

vf_info_t vf_info_export = {
    "Export video to shared memory",
    "export",
    "Otvos Attila",
    "",
    export_open,
    &vf_opts
};
