/*
 *  Import video frame from shmem stream.
 *
 *  Based on vf_bmovl.c
 *
 *  Copyright (C) 2007 Attila Ötvös <attila@onebithq.com>
 *
 *  MPlayer is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  MPlayer is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with MPlayer; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <fcntl.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "img_format.h"
#include "mp_image.h"
#include "vf.h"
#include "vf_scale.h"
#include "libswscale/swscale.h"

#include "mp_msg.h"
#include "libavutil/common.h"

#include "libvo/fastmemcpy.h"

#include "m_option.h"
#include "m_struct.h"

#include "shmem_reader.h"
#include "img_mixer.h"
#include "dynamicfilter.h"

// limited mpi size to sws_scaler
#define MAX_MPI_WIDTH       1920
#define MAX_MPI_HEIGHT      1024

#define ROUND_NONE          0
#define ROUND_DOWN          1
#define ROUND_UP            2
#define INT_NOSET           (int)(1<<31)

#define CMD_ON              0
#define CMD_FRAME_X         1
#define CMD_FRAME_Y         2
#define CMD_FRAME_W         3
#define CMD_FRAME_H         4
#define CMD_WINDOW_X        5
#define CMD_WINDOW_Y        6
#define CMD_WINDOW_W        7
#define CMD_WINDOW_H        8
#define CMD_ALPHA           9

static struct vf_priv_s {
    int id;
    int on;
    int frame_x, frame_y, frame_w, frame_h;
    int window_x, window_y, window_w, window_h;
    int alpha;

    int old_frame_w, old_frame_h;
    double start_pts, pts;
    double set_pts;
    int source_w, source_h;
    int scaled_x, scaled_y, scaled_w, scaled_h;
    int cut_x, cut_y, cut_w, cut_h;
    unsigned int frame_fmt, source_fmt;
    mp_image_t *src, *cut_src;
    mp_image_t *dst;
    void *video_reader;
    struct SwsContext *ctx;
    filter_cmd_t *cmd;
    filter_cmd_t *s_on, *e_on;
    filter_cmd_t *s_frame_x, *e_frame_x;
    filter_cmd_t *s_frame_y, *e_frame_y;
    filter_cmd_t *s_frame_w, *e_frame_w;
    filter_cmd_t *s_frame_h, *e_frame_h;
    filter_cmd_t *s_window_x, *e_window_x;
    filter_cmd_t *s_window_y, *e_window_y;
    filter_cmd_t *s_window_w, *e_window_w;
    filter_cmd_t *s_window_h, *e_window_h;
    filter_cmd_t *s_alpha, *e_alpha;
} const vf_priv_dflt = {
    0,                              // id
    -1,                             // on
    0, 0, 0, 0,                     // frame x,y,w,h
    0, 0, 0, 0,                     // window x,y,w,h
    -1,                             // alpha
    -1, -1,                         // old_frame w,h
    MP_NOPTS_VALUE, MP_NOPTS_VALUE, // start_pts, pts
    MP_NOPTS_VALUE,                 // set_pts
    0, 0,                           // source w,h
    0, 0, 0, 0,                     // scaled x,y,w,h
    0, 0, 0, 0,                     // cut x,y,w,h
    0, 0,                           // frame_fmt, source_fmt
    NULL, NULL,                     // src, cut_src
    NULL,                           // dst
    NULL,                           // video shmem reader priv
    NULL,                           // ctx
    NULL,                           // filter commands
    NULL,NULL,                      // dynamic on (start,end)
    NULL,NULL,                      // dynamic frame_x (start,end)
    NULL,NULL,                      // dynamic frame_y (start,end)
    NULL,NULL,                      // dynamic frame_w (start,end)
    NULL,NULL,                      // dynamic frame_h (start,end)
    NULL,NULL,                      // dynamic window_x (start,end)
    NULL,NULL,                      // dynamic window_y (start,end)
    NULL,NULL,                      // dynamic window_w (start,end)
    NULL,NULL,                      // dynamic window_h (start,end)
    NULL,NULL                       // dynamic alpha (start,end)
};

typedef struct vfcmd_priv_s {
    int on;
    int frame_x, frame_y, frame_w, frame_h;
    int window_x, window_y, window_w, window_h;
    int alpha;
} vfcmd_priv_t;

static vfcmd_priv_t const vfcmd_priv_dflt = {
    INT_NOSET,                                  // on
    INT_NOSET, INT_NOSET, INT_NOSET, INT_NOSET, // frame x,y,w,h
    INT_NOSET, INT_NOSET, INT_NOSET, INT_NOSET, // window x,y,w,h
    INT_NOSET                                   // alpha
};

static int query_format(struct vf_instance_s* vf, unsigned int fmt) {
    switch (fmt) {
    case IMGFMT_YV12:
    case IMGFMT_I420:
    case IMGFMT_IYUV:
    case IMGFMT_Y800:
    case IMGFMT_Y8:
    case IMGFMT_BGR24:
    case IMGFMT_RGB24:
    case IMGFMT_YUY2:
    case IMGFMT_UYVY:
        return VFCAP_CSP_SUPPORTED;
    }
    return 0;
}

/**
 * \brief set_scaler() set import frame properties
 * \param vf: video filter instances
 * \param force: calculate frame properties always
 * \return  0 - successfull
 * \return  1 - no change
 * \return -1 - error
 */
static int set_scaler(struct vf_instance_s* vf, int force, int d_width,
                      int d_height) {
    int posx, posy, sw, sh;
    int maxw=MAX_MPI_WIDTH;
    int maxh=MAX_MPI_HEIGHT;

    if (!vf->priv || !vf->priv->src) return 1;
    if (!force &&
            vf->priv->old_frame_w==vf->priv->frame_w &&
            vf->priv->old_frame_h==vf->priv->frame_h &&
            vf->priv->source_fmt==vf->priv->src->imgfmt &&
            vf->priv->source_w==vf->priv->src->width &&
            vf->priv->source_h==vf->priv->src->height &&
            vf->priv->ctx) {
        return 1;
    }
    if(!vf->priv->frame_w && !vf->priv->frame_h && vf->priv->source_fmt &&
       vf->priv->src) {
        vf->priv->frame_w=vf->priv->src->width;
        vf->priv->frame_h=vf->priv->src->height;
    }
    vf->priv->source_fmt = vf->priv->src->imgfmt;
    vf->priv->source_w = vf->priv->src->width;
    vf->priv->source_h = vf->priv->src->height;
    vf->priv->old_frame_w=vf->priv->frame_w;
    vf->priv->old_frame_h=vf->priv->frame_h;

    if (!vf->priv->window_w)
        vf->priv->window_w=FFMAX(0,d_width-vf->priv->window_x);
    if (!vf->priv->window_h)
        vf->priv->window_h=FFMAX(0,d_height-vf->priv->window_y);

    vf->priv->scaled_x = vf->priv->frame_x & ~1;
    vf->priv->scaled_y = vf->priv->frame_y & ~1;
    vf->priv->scaled_w = vf->priv->frame_w & ~1;
    vf->priv->scaled_h = vf->priv->frame_h & ~1;
    vf->priv->cut_x=0;
    vf->priv->cut_y=0;
    vf->priv->cut_w=0;
    vf->priv->cut_h=0;
    if (maxw<vf->priv->frame_w || maxh<vf->priv->frame_h) {
            posx=FFMAX(0,-vf->priv->frame_x);
            posy=FFMAX(0,-vf->priv->frame_y);
        vf->priv->cut_x=(posx*vf->priv->window_w/vf->priv->frame_w) & ~1;
        vf->priv->cut_y=(posy*vf->priv->window_h/vf->priv->frame_h) & ~1;
        vf->priv->cut_w=
            (vf->priv->window_w*vf->priv->window_w/vf->priv->frame_w) & ~1;
        vf->priv->cut_h=
            (vf->priv->window_h*vf->priv->window_h/vf->priv->frame_h) & ~1;
        vf->priv->scaled_w=
            (vf->priv->cut_w*vf->priv->frame_w/vf->priv->window_w) & ~1;
        vf->priv->scaled_h=
            (vf->priv->cut_h*vf->priv->frame_h/vf->priv->window_h) & ~1;
        vf->priv->scaled_x=-
            (vf->priv->window_x*vf->priv->window_w/vf->priv->frame_w) & ~1;
        vf->priv->scaled_y=-
            (vf->priv->window_y*vf->priv->window_h/vf->priv->frame_h) & ~1;
        vf->priv->scaled_x=-vf->priv->window_x;
        vf->priv->scaled_y=-vf->priv->window_y;
    }

    if (vf->priv->ctx) sws_freeContext(vf->priv->ctx);
    vf->priv->ctx = NULL;
    if (vf->priv->cut_w) sw=vf->priv->cut_w; else sw=vf->priv->source_w;
    if (vf->priv->cut_h) sh=vf->priv->cut_h; else sh=vf->priv->source_h;

    if (vf->priv->scaled_w>7 && vf->priv->scaled_h>7 && vf->priv->source_fmt
            && sw && sh && vf->priv->scaled_w<=maxw && vf->priv->scaled_h<=maxh)
        vf->priv->ctx = sws_getContextFromCmdLine(sw, sh,
                        vf->priv->source_fmt,
                        vf->priv->scaled_w,
                        vf->priv->scaled_h,
                        vf->priv->frame_fmt);

    if (vf->priv->dst) free_mp_image(vf->priv->dst);
    vf->priv->dst=NULL;
    if (vf->priv->cut_src) free_mp_image(vf->priv->cut_src);
    vf->priv->cut_src=NULL;
    if (!vf->priv->ctx)
        return 0;
    return 1;
}

static int get_cmd_var(vfcmd_priv_t *cmd_priv, int varid) {
    switch (varid) {
    case CMD_ON:
        return cmd_priv->on;
    case CMD_FRAME_X:
        return cmd_priv->frame_x;
    case CMD_FRAME_Y:
        return cmd_priv->frame_y;
    case CMD_FRAME_W:
        return cmd_priv->frame_w;
    case CMD_FRAME_H:
        return cmd_priv->frame_h;
    case CMD_WINDOW_X:
        return cmd_priv->window_x;
    case CMD_WINDOW_Y:
        return cmd_priv->window_y;
    case CMD_WINDOW_W:
        return cmd_priv->window_w;
    case CMD_WINDOW_H:
        return cmd_priv->window_h;
    case CMD_ALPHA:
        return cmd_priv->alpha;
    }
    return INT_NOSET;
}

static void init_dynvar(struct vf_instance_s* vf,
                        filter_cmd_t **start,
                        filter_cmd_t ** end,
                        int varid) {
    filter_cmd_t *cmd = vf->priv->cmd;
    int var;

    while (cmd) {
        var=get_cmd_var(cmd->opts,varid);
        if ((*start) && !(*end) && var!=INT_NOSET) {
            *end=cmd;
            return;
        }
        if (!(*start) && var!=INT_NOSET) *start=cmd;
        cmd=cmd->next;
    }
}

static void init_dynvars(struct vf_instance_s* vf) {
    if(vf->priv->start_pts<0)
        vf->priv->start_pts=0.0;
    if (!vf->priv->cmd) {
        if(vf->priv->on==-1)
            set_shmemreader_mode(vf->priv->video_reader,vf->priv->start_pts,
                            IMPORT_MODE_ON);
        else
            set_shmemreader_mode(vf->priv->video_reader,vf->priv->start_pts,
                            vf->priv->on);
        return;
    }
    init_dynvar(vf,&vf->priv->s_on,&vf->priv->e_on, CMD_ON);
    init_dynvar(vf,&vf->priv->s_frame_x,&vf->priv->e_frame_x, CMD_FRAME_X);
    init_dynvar(vf,&vf->priv->s_frame_y,&vf->priv->e_frame_y, CMD_FRAME_Y);
    init_dynvar(vf,&vf->priv->s_frame_w,&vf->priv->e_frame_w, CMD_FRAME_W);
    init_dynvar(vf,&vf->priv->s_frame_h,&vf->priv->e_frame_h, CMD_FRAME_H);
    init_dynvar(vf,&vf->priv->s_window_x,&vf->priv->e_window_x, CMD_WINDOW_X);
    init_dynvar(vf,&vf->priv->s_window_y,&vf->priv->e_window_y, CMD_WINDOW_Y);
    init_dynvar(vf,&vf->priv->s_window_w,&vf->priv->e_window_w, CMD_WINDOW_W);
    init_dynvar(vf,&vf->priv->s_window_h,&vf->priv->e_window_h, CMD_WINDOW_H);
    init_dynvar(vf,&vf->priv->s_alpha,&vf->priv->e_alpha, CMD_ALPHA);
    if(!vf->priv->s_on) {
        if(vf->priv->on==-1)
            set_shmemreader_mode(vf->priv->video_reader,vf->priv->start_pts,
                            IMPORT_MODE_ON);
        else
            set_shmemreader_mode(vf->priv->video_reader,vf->priv->start_pts,
                            vf->priv->on);
        return;
    }
    set_shmemreader_mode(vf->priv->video_reader,vf->priv->s_on->pts,
                    get_cmd_var(vf->priv->s_on->opts,CMD_ON));
}

static void cmd_to_dynvar(double pts, filter_cmd_t **start,
                          filter_cmd_t **end, int varid) {
    filter_cmd_t *cmd = *end;
    int var;

    if (!cmd) return;
    if (cmd->pts>=pts) return;
    *start=*end;
    cmd=(*end)->next;
    while (cmd) {
        var=get_cmd_var(cmd->opts,varid);
        if (var!=INT_NOSET) {
            *end=cmd;
            return;
        }
        cmd=cmd->next;
    }
    *end=NULL;
}

static void cmd_to_dynvars(struct vf_instance_s* vf, double pts) {
    cmd_to_dynvar(pts,&vf->priv->s_frame_x,&vf->priv->e_frame_x,CMD_FRAME_X);
    cmd_to_dynvar(pts,&vf->priv->s_frame_y,&vf->priv->e_frame_y,CMD_FRAME_Y);
    cmd_to_dynvar(pts,&vf->priv->s_frame_w,&vf->priv->e_frame_w,CMD_FRAME_W);
    cmd_to_dynvar(pts,&vf->priv->s_frame_h,&vf->priv->e_frame_h,CMD_FRAME_H);
    cmd_to_dynvar(pts,&vf->priv->s_window_x,&vf->priv->e_window_x,CMD_WINDOW_X);
    cmd_to_dynvar(pts,&vf->priv->s_window_y,&vf->priv->e_window_y,CMD_WINDOW_Y);
    cmd_to_dynvar(pts,&vf->priv->s_window_w,&vf->priv->e_window_w,CMD_WINDOW_W);
    cmd_to_dynvar(pts,&vf->priv->s_window_h,&vf->priv->e_window_h,CMD_WINDOW_H);
    cmd_to_dynvar(pts,&vf->priv->s_alpha,&vf->priv->e_alpha,CMD_ALPHA);
}

static void cmd_to_on(struct vf_instance_s* vf, double pts) {
    filter_cmd_t *cmd;
    int var;

    if(!vf->priv->s_on) return;
    if(vf->priv->s_on->pts!=vf->priv->set_pts) {
        var=get_cmd_var(vf->priv->s_on->opts,CMD_ON);
        set_shmemreader_mode(vf->priv->video_reader,vf->priv->s_on->pts,var);
        vf->priv->set_pts=vf->priv->s_on->pts;
        }
    if(vf->priv->s_on->pts>pts) return;
    cmd=vf->priv->s_on->next;
    while (cmd) {
        var=get_cmd_var(cmd->opts,CMD_ON);
        if (var!=INT_NOSET) {
            vf->priv->s_on=cmd;
            return;
        }
        cmd=cmd->next;
    }
}

static int dynvar_to_param(double pts, int *var, filter_cmd_t **start,
                           filter_cmd_t **end, int varid, int roundmode) {
    double rate,ds,de,dr;
    int new, ret = 0;

    while (1) {
        if (!(*start)) return 0;
        if ((*start)->pts>pts) return 0;
        if (!(*end) || ((*start)->pts==(*end)->pts &&
            (*start)->type!=FILTER_CMD_STOP)) {
            if (*var==get_cmd_var((*start)->opts,varid))
                return 0;
            *var=get_cmd_var((*start)->opts,varid);
            return 1;
        }
        if ((*end)->type==FILTER_CMD_STOP) {
            if (*var!=get_cmd_var((*start)->opts,varid)) {
                *var=get_cmd_var((*start)->opts,varid);
                ret=1;
            }
            if ((*start)->pts!=(*end)->pts)
                return ret;
            *start=*end;
            *end=(*end)->next;
            continue;
        }
        break;
    }
    rate=(pts-(*start)->pts)/((*end)->pts-(*start)->pts);
    if (rate>1.0) rate=1.0;
    ds=get_cmd_var((*start)->opts,varid);
    de=get_cmd_var((*end)->opts,varid);
    dr=(de-ds)*rate+ds;
    new=round(dr);
    switch (roundmode) {
    case ROUND_DOWN:
        new &= ~1;
        break;
    case ROUND_UP:
        new = (new+1)&~1;
        break;
    }
    if (*var==new) return ret;
    *var=new;
    return 1;
}

static int dynvars_to_param(struct vf_instance_s* vf, double pts) {
    int changed = 0;
    changed|=dynvar_to_param(pts,&vf->priv->frame_x,&vf->priv->s_frame_x,
                             &vf->priv->e_frame_x,CMD_FRAME_X,ROUND_UP);
    changed|=dynvar_to_param(pts,&vf->priv->frame_y,&vf->priv->s_frame_y,
                             &vf->priv->e_frame_y,CMD_FRAME_Y,ROUND_UP);
    changed|=dynvar_to_param(pts,&vf->priv->frame_w,&vf->priv->s_frame_w,
                             &vf->priv->e_frame_w,CMD_FRAME_W,ROUND_DOWN);
    changed|=dynvar_to_param(pts,&vf->priv->frame_h,&vf->priv->s_frame_h,
                             &vf->priv->e_frame_h,CMD_FRAME_H,ROUND_DOWN);
    changed|=dynvar_to_param(pts,&vf->priv->window_x,&vf->priv->s_window_x,
                             &vf->priv->e_window_x,CMD_WINDOW_X,ROUND_UP);
    changed|=dynvar_to_param(pts,&vf->priv->window_y,&vf->priv->s_window_y,
                             &vf->priv->e_window_y,CMD_WINDOW_Y,ROUND_UP);
    changed|=dynvar_to_param(pts,&vf->priv->window_w,&vf->priv->s_window_w,
                             &vf->priv->e_window_w,CMD_WINDOW_H,ROUND_DOWN);
    changed|=dynvar_to_param(pts,&vf->priv->window_h,&vf->priv->s_window_h,
                             &vf->priv->e_window_h,CMD_WINDOW_W,ROUND_DOWN);
    changed|=dynvar_to_param(pts,&vf->priv->alpha,&vf->priv->s_alpha,
                             &vf->priv->e_alpha,CMD_ALPHA,ROUND_NONE);
    return changed;
}

static int config(struct vf_instance_s* vf,
        int width, int height, int d_width, int d_height,
        unsigned int flags, unsigned int outfmt) {

    vf->priv->frame_fmt = outfmt;
    set_scaler(vf,1,d_width,d_height);
    return vf_next_config(vf, width, height, d_width, d_height, flags, outfmt);
}

static void uninit(struct vf_instance_s *vf) {
    if (vf->priv) {
        if (vf->priv->video_reader) close_video_shmemreader(vf->priv->video_reader);
        if (vf->priv->dst) free_mp_image(vf->priv->dst);
        if (vf->priv->cut_src) free_mp_image(vf->priv->cut_src);
        if (vf->priv->ctx) sws_freeContext(vf->priv->ctx);
        free(vf->priv);
    }
}

/**
 * \brief put_image() main import function
 * \param vf   : vf pointer
 * \param mpi  : input image
 * \param pts  : pts of frame
 * \return     : output image
 */
static int put_image(struct vf_instance_s* vf, mp_image_t* mpi, double pts) {
    mp_image_t* dmpi;

    dmpi = vf_get_image(vf->next, mpi->imgfmt, MP_IMGTYPE_TEMP,
                        MP_IMGFLAG_ACCEPT_STRIDE |
                        MP_IMGFLAG_PREFER_ALIGNED_STRIDE,
                        mpi->w, mpi->h);

    memcpy_pic( dmpi->planes[0], mpi->planes[0], mpi->width, mpi->height,
                dmpi->stride[0], mpi->stride[0] );
    memcpy_pic( dmpi->planes[1], mpi->planes[1], mpi->chroma_width,
                mpi->chroma_height, dmpi->stride[1], mpi->stride[1] );
    memcpy_pic( dmpi->planes[2], mpi->planes[2], mpi->chroma_width,
                mpi->chroma_height, dmpi->stride[2], mpi->stride[2] );

    if (pts==MP_NOPTS_VALUE) pts=vf->priv->pts;
    cmd_to_dynvars(vf,pts);
    cmd_to_on(vf,pts);
    if (!(vf->priv->src=(mp_image_t*)
                        get_next_shmemreader_frame(vf->priv->video_reader, pts)))
        return vf_next_put_image(vf, dmpi, MP_NOPTS_VALUE);
    if (!set_scaler(vf,dynvars_to_param(vf,pts),dmpi->width,dmpi->height))
        return vf_next_put_image(vf, dmpi, MP_NOPTS_VALUE);
    if (!vf->priv->dst)
        vf->priv->dst = alloc_mpi((vf->priv->scaled_w+15)&~15,
                                  (vf->priv->scaled_h+15)&~15,
                                  vf->priv->frame_fmt);
    if (!vf->priv->source_fmt)
        return vf_next_put_image(vf, dmpi, MP_NOPTS_VALUE);
    if (vf->priv->cut_w && vf->priv->cut_h) {
        if (!vf->priv->cut_src)
            vf->priv->cut_src = alloc_mpi((vf->priv->cut_w+15)&~15,
                                          (vf->priv->cut_h+15)&~15,
                                          vf->priv->source_fmt);
        if (!vf->priv->cut_src)
            return vf_next_put_image(vf, dmpi, MP_NOPTS_VALUE);
        cut_image(vf->priv->src,vf->priv->cut_src,vf->priv->cut_x,
                  vf->priv->cut_y,vf->priv->cut_w,vf->priv->cut_h);
    } else {
        if (vf->priv->cut_src) {
            free_mp_image(vf->priv->cut_src);
            vf->priv->cut_src=NULL;
        }
    }

    if (vf->priv->cut_src)
        sws_scale_ordered(vf->priv->ctx,
                          vf->priv->cut_src->planes,
                          vf->priv->cut_src->stride,
                          0,
                          vf->priv->cut_src->h,
                          vf->priv->dst->planes,
                          vf->priv->dst->stride);
    else
        sws_scale_ordered(vf->priv->ctx,
                          vf->priv->src->planes,
                          vf->priv->src->stride,
                          0,
                          vf->priv->src->h,
                          vf->priv->dst->planes,
                          vf->priv->dst->stride);

    mp_image_mixer(dmpi, vf->priv->dst,
                   vf->priv->window_x, vf->priv->window_y,
                   vf->priv->window_w, vf->priv->window_h,
                   vf->priv->scaled_x, vf->priv->scaled_y,
                   vf->priv->scaled_w, vf->priv->scaled_h,
                   vf->priv->alpha);
    return vf_next_put_image(vf, dmpi, MP_NOPTS_VALUE);
} // put_image

static int control(struct vf_instance_s* vf, int request, void* data) {

    switch(request){
    case VFCTRL_SEND_PTS:
        vf->priv->pts=*(double*)data;
        break;
    }
    return vf_next_control(vf, request, data);
}


static int vf_open(vf_instance_t* vf, char* args) {
    vf->config = config;
    vf->control = control;
    vf->put_image = put_image;
    vf->query_format = query_format;
    vf->uninit = uninit;

    if (!(vf->priv->video_reader=open_video_shmemreader(vf->priv->id)))
        return 0;

    vf->priv->frame_x &= ~1;
    vf->priv->frame_y &= ~1;
    if (vf->priv->frame_w==0) vf->priv->frame_w = vf->priv->source_w;
    if (vf->priv->frame_h==0) vf->priv->frame_h = vf->priv->source_h;
    vf->priv->scaled_w = vf->priv->frame_w;
    vf->priv->scaled_h = vf->priv->frame_h;
    vf->priv->old_frame_w = vf->priv->frame_w;
    vf->priv->old_frame_h = vf->priv->frame_h;
    vf->priv->scaled_w = (vf->priv->scaled_w+1)&~1;
    vf->priv->scaled_h = (vf->priv->scaled_h+1)&~1;
    vf->priv->cmd=get_filter_cmd(vf->priv->id,
                                 FILTER_CMD_TYPE_VIDEO,
                                 vf->info->name);
    init_dynvars(vf);
    return 1;
}

#define ST_OFF(f) M_ST_OFF(struct vf_priv_s,f)
static m_option_t vf_opts_fields[] = {
    {"id", ST_OFF(id), CONF_TYPE_INT, 0 , 0, 0, NULL},
    {"start-pts", ST_OFF(start_pts), CONF_TYPE_DOUBLE, M_OPT_MIN,0 , 0, NULL},
    {"on", ST_OFF(on), CONF_TYPE_INT, M_OPT_MIN,0 , 0, NULL},
    {"alpha", ST_OFF(alpha), CONF_TYPE_INT, M_OPT_MIN,0 , 0, NULL},
    {"frame-x", ST_OFF(frame_x), CONF_TYPE_INT, 0 ,0 , 0, NULL},
    {"frame-y", ST_OFF(frame_y), CONF_TYPE_INT, 0 ,0 , 0, NULL},
    {"frame-w", ST_OFF(frame_w), CONF_TYPE_INT, M_OPT_MIN,0 , 0, NULL},
    {"frame-h", ST_OFF(frame_h), CONF_TYPE_INT, M_OPT_MIN,0 , 0, NULL},
    {"window-x", ST_OFF(window_x), CONF_TYPE_INT, M_OPT_MIN,0 , 0, NULL},
    {"window-y", ST_OFF(window_y), CONF_TYPE_INT, M_OPT_MIN,0 , 0, NULL},
    {"window-w", ST_OFF(window_w), CONF_TYPE_INT, M_OPT_MIN,0 , 0, NULL},
    {"window-h", ST_OFF(window_h), CONF_TYPE_INT, M_OPT_MIN,0 , 0, NULL},
    { NULL, NULL, 0, 0, 0, 0,  NULL }
};

#undef ST_OFF
#define ST_OFF(f) M_ST_OFF(struct vfcmd_priv_s,f)
static m_option_t vfcmd_opts_fields[] = {
    {"on", ST_OFF(on), CONF_TYPE_INT, M_OPT_MIN, 0 , 0, NULL},
    {"alpha", ST_OFF(alpha), CONF_TYPE_INT, M_OPT_MIN,0 , 0, NULL},
    {"frame-x", ST_OFF(frame_x), CONF_TYPE_INT, 0 ,0 , 0, NULL},
    {"frame-y", ST_OFF(frame_y), CONF_TYPE_INT, 0 ,0 , 0, NULL},
    {"frame-w", ST_OFF(frame_w), CONF_TYPE_INT, M_OPT_MIN,0 , 0, NULL},
    {"frame-h", ST_OFF(frame_h), CONF_TYPE_INT, M_OPT_MIN,0 , 0, NULL},
    {"window-x", ST_OFF(window_x), CONF_TYPE_INT, M_OPT_MIN,0 , 0, NULL},
    {"window-y", ST_OFF(window_y), CONF_TYPE_INT, M_OPT_MIN,0 , 0, NULL},
    {"window-w", ST_OFF(window_w), CONF_TYPE_INT, M_OPT_MIN,0 , 0, NULL},
    {"window-h", ST_OFF(window_h), CONF_TYPE_INT, M_OPT_MIN,0 , 0, NULL},
    { NULL, NULL, 0, 0, 0, 0,  NULL }
};

static m_struct_t vfcmd_opts = {
    "import",
    sizeof(struct vfcmd_priv_s),
    &vfcmd_priv_dflt,
    vfcmd_opts_fields
};

dfcmd_info_t vfcmd_info_import = {
    "Video import filter from shmem stream",
    "import",
    "Otvos Attila",
    "",
    &vfcmd_opts
};


static m_struct_t vf_opts = {
    "import",
    sizeof(struct vf_priv_s),
    &vf_priv_dflt,
    vf_opts_fields
};

vf_info_t vf_info_import = {
    "Video import filter from shmem stream",
    "import",
    "Otvos Attila",
    "",
    vf_open,
    &vf_opts
};
