/*
 *  Demuxer to shared memory.
 *
 *  Copyright (C) 2007 Attila Ötvös <attila@onebithq.com>
 *
 *  MPlayer is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  MPlayer is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with MPlayer; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>

#include "m_option.h"

#include "stream/stream.h"
#include "stream/stream_shmem.h"
#include "stream/shmem_common.h"
#include "demuxer.h"
#include "stheader.h"

#include "libmpcodecs/img_format.h"

extern int demuxer_type;
extern double force_fps;

typedef struct demux_shmem_priv_st {
    struct shmem_video_st video;
    struct shmem_audio_st audio;
    int shmem_type;
    int realtime;
    double pts;
    double dpts;
    unsigned long long int now;
    char* video_data;
    char* audio_data;
    int next_packet;
    int video_pos;
    int audio_pos;
} demux_shmem_priv_t;

static int simple_read(stream_t *s, char *data, int size) {
    int r;
    while (size) {
        if(0>=(r=stream_read(s,data,size)))
            return 1;
        size-=r;
        data+=r;
    }
    return 0;
}

static void real_time_sleep(demux_shmem_priv_t *priv) {
    unsigned long long int now, corr;
    float pts_corr = 0.0f;

    if (priv->dpts > priv->pts) {
        pts_corr = priv->dpts-priv->pts;
        if (priv->dpts-priv->pts > 10.0f)
            now = (10.0f)*1000000.0f;
        else
            now = (priv->dpts-priv->pts)*1000000.0f;
        usleep(now);
    }
    now = GetTimer();
    if (priv->now == 0) priv->now = now;
    if (priv->now>now) corr = 0; else corr = now-priv->now;
    priv->pts += corr/1000000.0f;
    priv->now = now;
}

static int new_shmem_sh(demuxer_t* demuxer, demux_shmem_priv_t* priv) {
    sh_video_t* sh_video;
    sh_audio_t* sh_audio;
    WAVEFORMATEX* w;
    unsigned int packet_id;
    int total;
    char *ptr;

    packet_id=stream_read_dword_le(demuxer->stream);
    switch (packet_id) {
    case SHMEM_VIDEO_PACKET:
        if(priv->video_data && priv->next_packet)
            return 1;
        if(priv->video_data) {
            priv->next_packet=packet_id;
            return 0;
        }
        priv->video.id=packet_id;
        total=sizeof(shmem_video_t);
        ptr=(char*)&priv->video;
        total-=sizeof(int);
        ptr+=sizeof(int);
        if(simple_read(demuxer->stream,ptr,total))
            return 1;
        sh_video = new_sh_video(demuxer,0);
        sh_video->format=priv->video.format;
        sh_video->fps=25.0f;                                // FIXME
        sh_video->frametime=1.0/sh_video->fps;
        sh_video->disp_w=priv->video.width;
        sh_video->disp_h=priv->video.height;
        sh_video->i_bps=sh_video->fps*priv->video.size;

        demuxer->movi_start = demuxer->stream->start_pos;
        demuxer->movi_end = demuxer->stream->end_pos;

        demuxer->video->sh = sh_video;
        demuxer->priv=priv;
        sh_video->ds = demuxer->video;
        total=priv->video.size;
        priv->video_data=malloc(total);
        ptr=priv->video_data;
        if(simple_read(demuxer->stream,ptr,total))
            return 1;
        return 0;
    case  SHMEM_AUDIO_PACKET:
        if(priv->audio_data && priv->next_packet)
            return 1;
        if(priv->audio_data) {
            priv->next_packet=packet_id;
            return 0;
        }
        priv->audio.id=packet_id;
        total=sizeof(shmem_audio_t);
        ptr=(char*)&priv->audio;
        total-=sizeof(int);
        ptr+=sizeof(int);
        if(simple_read(demuxer->stream,ptr,total))
            return 1;
        sh_audio = new_sh_audio(demuxer,0);
        sh_audio->wf = w = malloc(sizeof(WAVEFORMATEX));
//        w->wFormatTag = sh_audio->format = priv->audio.format;
        w->wFormatTag = sh_audio->format = 0x01;                        // FIXME
        w->nChannels = sh_audio->channels = priv->audio.channels;
        w->nSamplesPerSec = sh_audio->samplerate = priv->audio.samplerate;
        if (priv->audio.bitrate > 999)
            w->nAvgBytesPerSec = priv->audio.bitrate/8;
        else if (priv->audio.bitrate > 0)
            w->nAvgBytesPerSec = priv->audio.bitrate*125;
        else
            w->nAvgBytesPerSec = priv->audio.samplerate*priv->audio.samplesize*
                                 priv->audio.channels;
        w->nBlockAlign = priv->audio.channels*priv->audio.samplesize;
        sh_audio->samplesize = priv->audio.samplesize;
        w->wBitsPerSample = 8*priv->audio.samplesize;
        w->cbSize = 0;

        demuxer->movi_start = demuxer->stream->start_pos;
        demuxer->movi_end = demuxer->stream->end_pos;

        demuxer->audio->sh = sh_audio;
        sh_audio->ds = demuxer->audio;
        total=priv->audio.size;
        priv->audio_data=malloc(total);
        ptr=priv->audio_data;
        if(simple_read(demuxer->stream,ptr,total))
            return 1;
        return 0;
    }
    return 1;
}

static demuxer_t* demux_shmem_open(demuxer_t* demuxer) {
    demux_shmem_priv_t *priv;

    priv=calloc(1,sizeof(demux_shmem_priv_t));
    if(1!=stream_control(demuxer->stream, STREAM_CTRL_GET_TYPE,
                         &priv->shmem_type)) {
        free(priv);
        return NULL;
    }
    if(priv->shmem_type!=SHMEM_TYPE_RAWDATA &&
       priv->shmem_type!=SHMEM_TYPE_PIPE) {
        free(priv);
        return NULL;
    }
    if(new_shmem_sh(demuxer,priv)) {
        free(priv);
        return NULL;
    }
    if(priv->shmem_type==SHMEM_TYPE_RAWDATA)
        return demuxer;
    if(new_shmem_sh(demuxer,priv)) {
        free(priv);
        return NULL;
    }
    return demuxer;
}

static int demux_shmem_fill_buffer(demuxer_t* demuxer, demux_stream_t *ds) {
    sh_video_t* sh_video = NULL;
    sh_audio_t* sh_audio = NULL;
    demux_packet_t* dp;
    demux_shmem_priv_t *priv = demuxer->priv;
    unsigned int packet_id=0;
    int total;
    char *ptr;

    if(demuxer->stream->eof) return 0;
    if(demuxer->video)
        sh_video = demuxer->video->sh;
    if(demuxer->audio)
        sh_audio = demuxer->video->sh;

    if(priv->video_data && demuxer->video) {
        dp=new_demux_packet(priv->video.size);
        memcpy(dp->buffer,priv->video_data,priv->video.size);
        free(priv->video_data);
        priv->video_data=NULL;
        resize_demux_packet(dp, priv->video.size);
        dp->pts=priv->video.pts;
        dp->pos=priv->video_pos;
        dp->flags=0x10;
        ds_add_packet(demuxer->video,dp);
        priv->video_pos++;
    }
    if(priv->audio_data && demuxer->audio) {
        dp=new_demux_packet(priv->audio.size);
        memcpy(dp->buffer,priv->audio_data,priv->audio.size);
        free(priv->audio_data);
        priv->audio_data=NULL;
        resize_demux_packet(dp, priv->audio.size);
        dp->pts=priv->audio.pts;
        dp->pos=priv->audio_pos;
        dp->flags=0x10;
        ds_add_packet(demuxer->audio,dp);
        priv->audio_pos++;
    }
    if(priv->next_packet) {
        packet_id=priv->next_packet;
        priv->next_packet=0;
    } else
        packet_id=stream_read_dword_le(demuxer->stream);
    switch (packet_id) {
    case SHMEM_VIDEO_PACKET:
        if(!demuxer->video)
            return 0;
        priv->video.id=packet_id;
        total=sizeof(shmem_video_t);
        ptr=(char*)&priv->video;
        total-=sizeof(int);
        ptr+=sizeof(int);
        if(simple_read(demuxer->stream,ptr,total))
            return 0;
        sh_video->format=priv->video.format;
        sh_video->disp_w=priv->video.width;
        sh_video->disp_h=priv->video.height;
        if(priv->shmem_type==SHMEM_TYPE_RAWDATA) {
	    if(force_fps==0.0)
        	priv->dpts+=(float)1/25.0;
		else
        	priv->dpts+=(float)1/force_fps;
	    if(priv->realtime)
		real_time_sleep(priv);
            priv->video.pts=priv->dpts;
        }
        ds_read_packet(demuxer->video,demuxer->stream,priv->video.size,
                       priv->video.pts,priv->video_pos,0x10);
        priv->video_pos++;
        return 1;
    case SHMEM_AUDIO_PACKET:
        if(!demuxer->audio)
            return 0;
        priv->audio.id=packet_id;
        total=sizeof(shmem_audio_t);
        ptr=(char*)&priv->audio;
        total-=sizeof(int);
        ptr+=sizeof(int);
        if(simple_read(demuxer->stream,ptr,total))
            return 0;
        ds_read_packet(demuxer->audio,demuxer->stream,priv->audio.size,
                       priv->audio.pts,priv->audio_pos,0x10);
        priv->audio_pos++;
        return 1;
    }
    return 0;
}

static void demux_shmem_close(demuxer_t *demuxer) {
    demux_shmem_priv_t *priv = demuxer->priv;

    if(priv) {
        if(priv->video_data) free(priv->video_data);
        if(priv->audio_data) free(priv->audio_data);
        free(priv);
    }
}

static int demux_shmem_control(demuxer_t *demuxer, int cmd, void *arg) {
    demux_shmem_priv_t *priv = demuxer->priv;
  
    switch (cmd) {
    case DEMUXER_CTRL_SETREALTIME:
	priv->realtime=1;
	return DEMUXER_CTRL_OK;
    default:
	return DEMUXER_CTRL_NOTIMPL;
    }
}

demuxer_desc_t demuxer_desc_shmem = {
    "Shmem video demuxer",
    "shmem",
    "shmem",
    "?",
    "",
    DEMUXER_TYPE_SHMEM,
    0, // no autodetect
    NULL,
    demux_shmem_fill_buffer,
    demux_shmem_open,
    demux_shmem_close,
    NULL,
    demux_shmem_control
};
