/*
 *  Muxer to shared memory
 *
 *  Based on vf.c
 *
 *  Copyright (C) 2007 Attila Ötvös <attila@onebithq.com>
 *
 *  MPlayer is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  MPlayer is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with MPlayer; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>
#include <unistd.h>

#include "config.h"
#include "version.h"

#include "aviheader.h"
#include "ms_hdr.h"

#include "stream/stream.h"
#include "stream/shmem_common.h"
#include "stream/stream_shmem.h"
#include "libaf/af.h"
#include "muxer.h"

static muxer_stream_t* shmem_new_stream(muxer_t *muxer,int type){
    muxer_stream_t* s;
    if (!muxer) return NULL;
    s=malloc(sizeof(muxer_stream_t));
    memset(s,0,sizeof(muxer_stream_t));
    if (!s) return NULL; // no mem!?
    muxer->streams[muxer->avih.dwStreams]=s;
    s->type=type;
    s->id=muxer->avih.dwStreams;
    s->timer=0.0;
    s->size=0;
    s->muxer=muxer;
    switch(type){
    case MUXER_TYPE_VIDEO:
        s->ckid=mmioFOURCC(('0'+s->id/10),('0'+(s->id%10)),'d','c');
        s->h.fccType=streamtypeVIDEO;
        if (!muxer->def_v) muxer->def_v=s;
        break;
    case MUXER_TYPE_AUDIO:
        s->ckid=mmioFOURCC(('0'+s->id/10),('0'+(s->id%10)),'d','c');
        s->h.fccType=streamtypeAUDIO;
        break;
    }
    muxer->avih.dwStreams++;
    return s;
}

static void write_shmem_video_chunk(stream_t *stream,int len,void* data,
                                    float pts,muxer_stream_t *ms) {
    shmem_video_t *video = NULL;
    char *ptr;
    int total;
    int r;

    if (len<=0 || !data || !stream->type==STREAMTYPE_SHMEM)
        return;
    video=calloc(1,sizeof(shmem_video_t));
    video->id=SHMEM_VIDEO_PACKET;
    video->format=ms->imgfmt;
    video->width=ms->bih->biWidth;
    video->height=ms->bih->biHeight;
    video->aspect=ms->aspect;
    video->size=len;
    video->bytes=0;
    video->pts=pts;
    ptr=(char*)video;
    total=sizeof(shmem_video_t);
    while (total) {
        if(0>=(r=stream_write_buffer(stream, ptr, total))) {
            stream_at_eof=1;
            free(video);
            return;
        }
        total-=r;
        ptr+=r;
    }
    free(video);
    while (len) {
        if(0>=(r=stream_write_buffer(stream, data, len))) {
            stream_at_eof=1;
            return;
        }
        len-=r;
        data+=r;
    }
}

void write_shmem_audio_chunk(stream_t *stream,int len,
                             void* data, float pts,muxer_stream_t *ms) {
    shmem_audio_t *audio = NULL;
    char *ptr;
    int total;
    int r;

    if (len<=0 || !data || !stream->type==STREAMTYPE_SHMEM)
        return;
    audio=calloc(1,sizeof(shmem_audio_t));
    audio->id=SHMEM_AUDIO_PACKET;
    audio->format=(ms->wf->wBitsPerSample==8) ? AF_FORMAT_U8 : AF_FORMAT_S16_LE;
    audio->channels=ms->wf->nChannels;
    audio->samplerate=ms->h.dwRate;
    audio->samplesize=ms->h.dwSampleSize;
    audio->samplesize=(ms->wf->wBitsPerSample==8) ? 1 : 2;        // FIXME
    audio->bitrate=0;
    audio->size=len;
    audio->bytes=0;
    audio->pts=pts;
    ptr=(char*)audio;
    total=sizeof(shmem_audio_t);
    while (total) {
        if(0>=(r=stream_write_buffer(stream, ptr, total))) {
            stream_at_eof=1;
            free(audio);
            return;
        }
        total-=r;
        ptr+=r;
    }
    free(audio);
    while (len) {
        if(0>=(r=stream_write_buffer(stream, data, len))) {
            stream_at_eof=1;
            return;
        }
        len-=r;
        data+=r;
    }
}

static void shmem_write_chunk(muxer_stream_t *s,size_t len,
                              unsigned int flags, double dts, double pts) {
    muxer_t *muxer=s->muxer;

    // write out the chunk:
    if (s->type == MUXER_TYPE_VIDEO)
        write_shmem_video_chunk(muxer->stream,len,s->buffer,pts,s);
    if (s->type == MUXER_TYPE_AUDIO)
        write_shmem_audio_chunk(muxer->stream,len,s->buffer,pts,s);
}

static void shmem_write_header(muxer_t *muxer){
    return;
}

static void shmem_write_index(muxer_t *muxer){
    return;
}

int muxer_init_muxer_shmem(muxer_t *muxer){
    muxer->cont_new_stream = &shmem_new_stream;
    muxer->cont_write_chunk = &shmem_write_chunk;
    muxer->cont_write_header = &shmem_write_header;
    muxer->cont_write_index = &shmem_write_index;
    return 1;
}
