/*
 *  Reader to shmem stream.
 *
 *  Based on vf.c
 *
 *  Copyright (C) 2007 Attila Ötvös <attila@onebithq.com>
 *
 *  MPlayer is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  MPlayer is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with MPlayer; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>
#include <ctype.h>

#include "config.h"

#include "mp_msg.h"
#include "help_mp.h"

#include "m_option.h"
#include "m_struct.h"

#include "stream/stream.h"
#include "libmpdemux/demuxer.h"
#include "libmpdemux/stheader.h"
#include "libmpcodecs/img_format.h"
#include "libmpcodecs/mp_image.h"
#include "libaf/af.h"

#include "shmem_reader.h"

#define AUDIO_BUFFER_SIZE       0x100000
#define AUDIO_PTS_DIFF          1.0

#define READER_FLAG_VIDEO       1
#define READER_FLAG_AUDIO       2
#define READER_FLAG_START       4
#define READER_FLAG_PAUSE       8
#define READER_FLAG_EOF         16
#define READER_FLAG_ASTART      32

typedef struct reader_priv_s {
    int             id;
    stream_t        *stream;
    demuxer_t       *demuxer;
    sh_video_t      *sh_video;
    sh_audio_t      *sh_audio;
    int                    flag;
// video reader
    mp_image_t      *src0;
    mp_image_t      *src1;
    double          pts0;
    double          pts1;
    double          pts_offset;
    double          set_pts;
    double          set_apts;
    int             set_mode;
    double          frametime;
// audio reader
    unsigned char*  buffer;
    int                    bytes;
    double          apts;

    void*           next;
} reader_priv_t;

static reader_priv_t def_reader_priv = {
    0,               // id
    NULL,            // stream
    NULL,            // demuxer
    NULL,            // sh_video
    NULL,            // sh_audio
    0,               // flag

    NULL,            // src0
    NULL,            // src1
    MP_NOPTS_VALUE,  // pts0
    MP_NOPTS_VALUE,  // pts1
    MP_NOPTS_VALUE,  // pts_offset
    MP_NOPTS_VALUE,  // set_pts
    MP_NOPTS_VALUE,  // set_apts
    0,               // set_mode
    0.0,             // frametime

    NULL,            // buffer
    0,               // bytes
    MP_NOPTS_VALUE,  // apts

    NULL             // next
};

reader_priv_t *readers_priv = NULL;

static reader_priv_t* new_reader_priv(int id){
    reader_priv_t *tmp = readers_priv;
    while (tmp) {
        if (tmp->id==id) return tmp;
        if (!tmp->next) {
            tmp->next=malloc(sizeof(reader_priv_t));
            tmp=tmp->next;
            *tmp=def_reader_priv;
            tmp->id=id;
            return tmp;
        }
        tmp=tmp->next;
    }
    readers_priv=malloc(sizeof(reader_priv_t));
    *readers_priv=def_reader_priv;
    readers_priv->id=id;
    return readers_priv;
}

static reader_priv_t* get_reader_priv(int id){
    reader_priv_t *tmp = readers_priv;
    while (tmp) if (tmp->id==id) return tmp; else tmp=tmp->next;
    return new_reader_priv(id);
    return NULL;
}

static void free_reader_priv(int id){
    reader_priv_t *tmp = readers_priv;
    reader_priv_t *ptmp = NULL;
    while (tmp)
            if (tmp->id==id) break;
        else
            {ptmp=tmp; tmp=tmp->next;}
    if (!tmp) return;
    if (ptmp) ptmp->next=tmp->next; else readers_priv=tmp->next;
    free(tmp);
}

static void* close_reader_priv(reader_priv_t *priv, int flag) {

    if(!priv) return NULL;
    priv->flag&=~flag;
    if(priv->flag) return NULL;
    if (priv->demuxer) {
        free_demuxer(priv->demuxer);
        priv->sh_audio=NULL;
        priv->sh_video=NULL;
    }
    priv->demuxer=NULL;
    if (priv->stream)
        free_stream(priv->stream);
    priv->stream=NULL;
    if (priv->buffer)
        free(priv->buffer);
    priv->buffer=NULL;
    if (priv->src0)
        free_mp_image(priv->src0);
    priv->src0=NULL;
    if (priv->src1)
        free_mp_image(priv->src1);
    priv->src1=NULL;
    free_reader_priv(priv->id);
    return NULL;
}

static void read_audio(reader_priv_t *reader_priv, double pts) {
    int total, len, droplen;
    char *rbuffer;
    unsigned char* start = NULL;

    if (!reader_priv) return;
    if (!reader_priv->sh_audio) return;
    if (!reader_priv->buffer) {
        reader_priv->buffer=malloc(AUDIO_BUFFER_SIZE);
        memset(reader_priv->buffer,0,AUDIO_BUFFER_SIZE);
        reader_priv->bytes=0;
    }
    if(!reader_priv->demuxer->audio->packs)
        return;
    total=AUDIO_BUFFER_SIZE-reader_priv->bytes;
    rbuffer=reader_priv->buffer+reader_priv->bytes;
    len=ds_get_packet_pts(reader_priv->demuxer->audio,&start,
                          &reader_priv->apts);
    if(len<0) len=0;
    if(len>total) {
        droplen=FFMIN(len,reader_priv->bytes);
        if (droplen) {
            reader_priv->bytes -= droplen;
            if (reader_priv->bytes)
                    memmove(reader_priv->buffer, reader_priv->buffer + droplen,
                            reader_priv->bytes);
        total=AUDIO_BUFFER_SIZE-reader_priv->bytes;
        rbuffer=reader_priv->buffer+reader_priv->bytes;
        }
    }
    len=FFMIN(len,total);
    if(len && start)
        memcpy(rbuffer,start,len);
    reader_priv->bytes+=len;
}

void* open_audio_shmemreader(int id) {
    reader_priv_t *reader_priv;
    char filename[64];
    int file_format;

    if(!(reader_priv = get_reader_priv(id)))
        return NULL;
    if (reader_priv->flag & READER_FLAG_AUDIO) {
        mp_msg(MSGT_VFILTER, MSGL_V,
               "Open audio stream is busy. ID: %d \n",
               reader_priv->id);
        return NULL;
    }
    reader_priv->flag|=READER_FLAG_AUDIO;
    if (!reader_priv->stream) {
        snprintf(filename,sizeof(filename),"shmem://%d",reader_priv->id);
        if (!(reader_priv->stream = open_stream(filename,0,&file_format)))
            return close_reader_priv(reader_priv,READER_FLAG_AUDIO);

        if(!(reader_priv->demuxer=demux_open(reader_priv->stream,file_format,
                                             0,0,0,filename)))
            return close_reader_priv(reader_priv,READER_FLAG_AUDIO);
        if(reader_priv->demuxer->video)
            reader_priv->sh_video=reader_priv->demuxer->video->sh;
        if(reader_priv->demuxer->audio)
            reader_priv->sh_audio=reader_priv->demuxer->audio->sh;
    }
    printf("Open #%d. shmem audio: %d Hz, %d ch, %d bytes/sec\n",
           reader_priv->id,
           reader_priv->sh_audio->samplerate,
           reader_priv->sh_audio->channels,
           reader_priv->sh_audio->wf->nAvgBytesPerSec);
    read_audio(reader_priv,MP_NOPTS_VALUE);
    return reader_priv;
}

void* get_shmemreader_audio_data(void *priv, int samples, double pts) {
    reader_priv_t *p = priv;
    af_data_t *filter_input;
    char *rbuffer, *ptr;
    unsigned char* start = NULL;
    int len, samplesize, minlen, offset=0;
    double timelen;

    if (!p) return NULL;
    if (!p->sh_audio) return NULL;
    if (!p->buffer) {
        p->buffer=malloc(AUDIO_BUFFER_SIZE);
        memset(p->buffer,0,AUDIO_BUFFER_SIZE);
        p->bytes=0;
    }
    samplesize=(p->sh_audio->wf->wBitsPerSample==8) ? 1 : 2;
    minlen=samples*samplesize*p->sh_audio->channels;
    timelen=(float)samples/(float)p->sh_audio->samplerate;
    if (pts<0.0) pts=0.0;
    if(p->set_apts!=MP_NOPTS_VALUE) {
        switch (p->set_mode) {
        case IMPORT_MODE_OFF:
        case IMPORT_MODE_PAUSE:
            if(p->set_apts<=pts) {
                offset=(p->set_apts-pts)*p->sh_audio->samplerate;
                offset=offset*samplesize*p->sh_audio->channels;
                p->flag&=~READER_FLAG_ASTART;
                p->set_apts=MP_NOPTS_VALUE;
            }
            break;
        case IMPORT_MODE_ON:
        case IMPORT_MODE_SKIPEOF:
            if(p->set_apts<=pts+timelen ||
               (p->set_apts<=pts && p->set_apts<timelen)) {
                offset=(p->set_apts-pts)*p->sh_audio->samplerate;
                offset=offset*samplesize*p->sh_audio->channels;
                p->flag|=READER_FLAG_ASTART;
                p->set_apts=MP_NOPTS_VALUE;
            }
            break;
        }
    }
    if(!(p->flag&READER_FLAG_ASTART)) return NULL;
    filter_input=malloc(sizeof(af_data_t));
    memset(filter_input,0,sizeof(af_data_t));
    filter_input->rate=p->sh_audio->samplerate;
    filter_input->nch=p->sh_audio->channels;
    filter_input->format=(p->sh_audio->wf->wBitsPerSample==8) ?
                          AF_FORMAT_U8 : AF_FORMAT_S16_LE;
    if(p->bytes<minlen) {
        len=ds_get_packet_pts(p->demuxer->audio,&start,&p->apts);
        if(len<0) len=0;
        rbuffer=p->buffer+p->bytes;
        if(len && start)
            memcpy(rbuffer,start,len);
        p->bytes+=len;
    }
    if(p->bytes<minlen) {
        free(filter_input);
        return NULL;
    }
    if(offset>0) {
        if (offset>minlen) {
            free(filter_input);
            return NULL;
        }
        filter_input->audio=calloc(1,minlen);
        ptr=filter_input->audio;
        ptr+=offset;
        minlen-=offset;
        memcpy(ptr,p->buffer,minlen);
        filter_input->len=minlen;
        af_fix_parameters(filter_input);
        return filter_input;
    }
    minlen+offset;
    if(minlen<=0) {
        free(filter_input);
        return NULL;
    }
    filter_input->len=minlen;
    filter_input->audio=malloc(minlen);
    memcpy(filter_input->audio,p->buffer,minlen);
    af_fix_parameters(filter_input);
    return filter_input;
}

void drop_shmemreader_audio_data(void *priv, void* filter_input, int len) {
    reader_priv_t *p = priv;
    af_data_t *af_data = filter_input;
    if (!p) return;
    if (!filter_input) return;
    len=min(len,p->bytes);
    if (len) {
        p->bytes -= len;
        if (p->bytes)
                memmove(p->buffer, p->buffer + len, p->bytes);
    }
    if(af_data) {
        if (af_data->audio)
            free(af_data->audio);
        free(af_data);
    }
}


void close_audio_shmemreader(void *priv) {
    reader_priv_t *p = priv;
    close_reader_priv(p,READER_FLAG_AUDIO);
}


void* open_video_shmemreader(int id) {
    reader_priv_t *reader_priv;
    char filename[64];
    int file_format;

    if(!(reader_priv = get_reader_priv(id)))
        return NULL;
    if (reader_priv->flag & READER_FLAG_VIDEO) {
        mp_msg(MSGT_VFILTER, MSGL_V,
               "Open video stream is busy. ID: %d \n",
               reader_priv->id);
        return NULL;
    }
    reader_priv->flag|=READER_FLAG_VIDEO;
    if (!reader_priv->stream) {
        snprintf(filename,sizeof(filename),"shmem://%d",id);
        if (!(reader_priv->stream = open_stream(filename,0,&file_format)))
            return close_reader_priv(reader_priv,READER_FLAG_VIDEO);

        if(!(reader_priv->demuxer=demux_open(reader_priv->stream,file_format,
                                             0,0,0,filename)))
            return close_reader_priv(reader_priv,READER_FLAG_VIDEO);
        if(reader_priv->demuxer->video)
            reader_priv->sh_video=reader_priv->demuxer->video->sh;
        if(reader_priv->demuxer->audio)
            reader_priv->sh_audio=reader_priv->demuxer->audio->sh;
    }

    printf("Open #%d. shmem video: %d x %d (%s)\n",
           reader_priv->id,
           reader_priv->sh_video->disp_w,
           reader_priv->sh_video->disp_h,
           vo_format_name(reader_priv->sh_video->format));
    return reader_priv;
}

/**
 * \brief read_frame() read image frame from insert stream
 * \param priv : reader_priv pointer
 * \param srv  : image pointer
 * \param pts  : pts of readed frame
 * \return     : readed image
 */
static mp_image_t* read_frame(reader_priv_t *priv,
                              mp_image_t *src,
                              double *pts,
                              double pts_offset) {
    unsigned char* start = NULL;
    int total, len;
    double spts;

    if(!priv->sh_video) return src;
    read_audio(priv,*pts);
    if(src && (src->width!=priv->sh_video->disp_w ||
               src->height!=priv->sh_video->disp_h ||
               src->imgfmt!=priv->sh_video->format)) {
        free_mp_image(src);
        src=NULL;
    }
    if(!src) src=alloc_mpi(priv->sh_video->disp_w,
                           priv->sh_video->disp_h,
                           priv->sh_video->format);

    total=src->width*src->height*src->bpp/8;;
    if (!total) {
        free_mp_image(src);
        src=NULL;
        if (pts) *pts=MP_NOPTS_VALUE;
        return src;
    }
    len=ds_get_packet_pts(priv->demuxer->video,&start, &spts);
    if(spts==MP_NOPTS_VALUE) spts=0.0;
    if(len!=total || !start) {
        free_mp_image(src);
        src=NULL;
        if (pts) *pts=MP_NOPTS_VALUE;
        return src;
    }
    memcpy(src->planes[0],start,len);
    if (!pts) return src;
    if(priv->pts_offset<0.0) {
        if(spts>0.0)
            priv->pts_offset=pts_offset-spts;
        else
            priv->pts_offset=pts_offset;
    }
    *pts=spts+priv->pts_offset;
    return src;
}

/**
 * \brief read_frames() read next image frame from shmem stream
 * \param priv : reader_priv_t pointer
 * \return     : 1 - succesful
 * \return     : 0 - error
 */
static int read_frames(reader_priv_t* priv, double pts_offset) {
    mp_image_t *tmp;
    if(priv->pts0==MP_NOPTS_VALUE) {
        priv->src0=read_frame(priv,priv->src0,&priv->pts0,pts_offset);
        if(!priv->src0) return 0;
    } else {
        priv->pts0=priv->pts1;
        tmp=priv->src0;
        priv->src0=priv->src1;
        priv->src1=tmp;
    }
    priv->src1=read_frame(priv,priv->src1,&priv->pts1,pts_offset);
    if(!priv->src1) {
        priv->flag|=READER_FLAG_EOF;
        return 0;
    }
    return 1;
}

void* get_next_shmemreader_frame(void *priv, double pts) {
    reader_priv_t *p = priv;
    mp_image_t *src = NULL;

    if (pts!=MP_NOPTS_VALUE) {
        if(p->set_pts<=pts && p->set_pts!=MP_NOPTS_VALUE) {
            switch (p->set_mode) {
            case IMPORT_MODE_OFF:
                p->flag&=~READER_FLAG_START;
                p->flag&=~READER_FLAG_PAUSE;
                break;
            case IMPORT_MODE_ON:
                p->flag|=READER_FLAG_START;
                p->flag&=~READER_FLAG_PAUSE;
                break;
            case IMPORT_MODE_PAUSE:
                p->flag&=~READER_FLAG_START;
                p->flag|=READER_FLAG_PAUSE;
                break;
            case IMPORT_MODE_SKIPEOF:
                p->flag|=READER_FLAG_START;
                p->flag|=READER_FLAG_PAUSE;
                break;
            }
        p->set_pts=MP_NOPTS_VALUE;
        }
    }
    if (!(p->flag&READER_FLAG_START)) {
        p->pts_offset=MP_NOPTS_VALUE;
        p->pts0=MP_NOPTS_VALUE;
        p->pts1=MP_NOPTS_VALUE;
        p->frametime=0.0;
        if(p->flag&READER_FLAG_PAUSE) {
            return p->src0;
        }
        return NULL;
    }
    if (pts==MP_NOPTS_VALUE) {
        src=read_frame(p,src,NULL,pts);
        return src;
    }
    while (!(p->flag&READER_FLAG_EOF)) {
        if (p->pts0!=MP_NOPTS_VALUE && p->pts1!=MP_NOPTS_VALUE &&
                    p->pts0<=pts && p->pts1>=pts) {
            p->frametime=p->pts1-p->pts0;
            if(pts-p->pts0<p->pts1-pts)
                    src=p->src0;
            else
                    src=p->src1;
            return src;
        }
        if (p->pts0!=MP_NOPTS_VALUE && p->pts1==MP_NOPTS_VALUE &&
                p->pts0>=pts && p->pts0+p->frametime<=pts) {
            src=p->src0;
            return src;
            }
        if (p->pts0!=MP_NOPTS_VALUE && p->pts0>pts) break;
        if (!read_frames(p,pts)) break;
    }
    if(p->flag&READER_FLAG_EOF && p->flag&READER_FLAG_PAUSE)
        return p->src0;
    return NULL;
}

void set_shmemreader_mode(void *priv, double pts, int mode) {
    reader_priv_t *p = priv;
    if(pts==MP_NOPTS_VALUE)
        return;
    p->set_pts=pts;
    p->set_apts=pts;
    p->set_mode=mode;
}

void close_video_shmemreader(void *priv) {
    reader_priv_t *p = priv;
    close_reader_priv(p,READER_FLAG_VIDEO);
}
