/*
 *  Special blank stream to video insert mode.
 *
 *  Copyright (C) 2007 Attila Ötvös <attila@onebithq.com>
 *
 *  MPlayer is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  MPlayer is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with MPlayer; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>

#include "stream.h"
#include "m_option.h"
#include "m_struct.h"
#include "libavutil/common.h"
#include "libmpdemux/demuxer.h"
#include "libmpcodecs/img_format.h"
#include "libmpcodecs/mp_image.h"
#include "libswscale/swscale.h"
#include "libmpcodecs/vf_scale.h"

extern int blank_v_format;
extern int blank_v_size_id;
extern int blank_v_width;
extern int blank_v_height;
extern int blank_v_imgsize;
extern int blank_v_aspect;
extern int blank_v_color;

static struct stream_blank_priv_s {
    char* color;
    int imgsize;
    int v_readptr;
    mp_image_t *img;
} stream_priv_dflts = {
    NULL, // color
    0,    // imgsize
    0,    // v_readptr
    NULL  // img
};


#define ST_OFF(f) M_ST_OFF(struct stream_blank_priv_s,f)
/// URL definition
static m_option_t stream_opts_fields[] = {
    {"hostname", ST_OFF(color), CONF_TYPE_STRING, 0, 0 ,0, NULL},
    { NULL, NULL, 0, 0, 0, 0,  NULL }
};
static struct m_struct_st stream_opts = {
    "blank",
    sizeof(struct stream_blank_priv_s),
    &stream_priv_dflts,
    stream_opts_fields
};

struct {
    int col;
    char *colname;
} bkcolors_table[] = {
        {0x000000,"black"},
        {0x0000ff,"blue"},
        {0x808080,"gray"},
        {0x008000,"green"},
        {0xff00ff,"fuchsia"},
        {0x00ff00,"lime"},
        {0x800000,"maroon"},
        {0x000080,"navy"},
        {0x808000,"olive"},
        {0x800080,"purple"},
        {0xff0000,"red"},
        {0xc0c0c0,"silver"},
        {0x008080,"teal"},
        {0xffffff,"white"},
        {0xffff00,"yellow"},
        {NULL,0}
};

static int fill_buffer(stream_t *s, char* buffer, int max_len){
    struct stream_blank_priv_s* p = (struct stream_blank_priv_s*)s->priv;
    int len;
    char *ptr;

    len=FFMIN(max_len,p->imgsize-p->v_readptr);
    ptr=p->img->planes[0]+p->v_readptr;
    memcpy(buffer,ptr,len);
    p->v_readptr+=len;
    if(p->v_readptr==p->imgsize)
        p->v_readptr=0;
    return len;
};

static void close_s(struct stream_st *s) {
    struct stream_blank_priv_s* p = (struct stream_blank_priv_s*)s->priv;

    if(p->img) free_mp_image(p->img);
    m_struct_free(&stream_opts,p);
}

static int open_s(stream_t *stream,int mode, void* opts, int* file_format) {
    struct stream_blank_priv_s* p = (struct stream_blank_priv_s*)opts;
    unsigned char col0, col1, col2, *pimg;
    char *colname;
    unsigned int col;
    struct SwsContext *ctx;
    mp_image_t *mpi;
    int i,j;

    if (mode!=STREAM_READ)
        return STREAM_UNSUPPORTED;

    switch(blank_v_size_id){
    case 1: blank_v_width=128; blank_v_height=96; break;
    case 2: blank_v_width=176; blank_v_height=144; break;
    case 3: blank_v_width=352; blank_v_height=288; break;
    case 4: blank_v_width=704; blank_v_height=576; break;
    case 5: blank_v_width=720; blank_v_height=576; break;
    case 6: blank_v_width=720; blank_v_height=480; break;
    case 7: blank_v_width=1408;blank_v_height=1152;break;
    case 8: blank_v_width=352; blank_v_height=240; break;
    }
    if(!blank_v_width || !blank_v_height){
      blank_v_width=720;
      blank_v_height=576;
    }
    if(!blank_v_format)
        blank_v_format = IMGFMT_I420;

    p->img=alloc_mpi(blank_v_width,blank_v_height,blank_v_format);
    blank_v_imgsize=p->imgsize=p->img->width*p->img->height*p->img->bpp/8;
    if(!p->img || !blank_v_format || !p->imgsize){
        mp_msg(MSGT_DEMUX,MSGL_ERR,
               "blank: img size not specified and unknown format!\n");
        return 0;
    }

    col=0;
    if(p->color)
        colname=p->color;
    else
        colname=blank_v_color;
    if(colname) {
        i = 0;
        sscanf(colname,"%x",&col);
        while(bkcolors_table[i].colname != NULL) {
            if( !strcasecmp( colname, bkcolors_table[i].colname ) ) {
                col=bkcolors_table[i].col;
                break;
            }
            i++;
        }
    }
    col0=(col>>16)&0xff;
    col1=(col>>8)&0xff;
    col2=col&0xff;
    mpi=alloc_mpi(blank_v_width,blank_v_height,IMGFMT_RGB24);
    for(i=0;i<blank_v_height;i++) {
        pimg=mpi->planes[0]+(mpi->stride[0]*i);
        for(j=0;j<blank_v_width;j++) {
            *pimg++=col0;
            *pimg++=col1;
            *pimg++=col2;
        }
    }
    ctx=sws_getContextFromCmdLine(blank_v_width, blank_v_height,
                        IMGFMT_RGB24,
                        blank_v_width,
                        blank_v_height,
                        blank_v_format);
    sws_scale_ordered(ctx,
                          mpi->planes,
                          mpi->stride,
                          0,
                          blank_v_height,
                          p->img->planes,
                          p->img->stride);
    sws_freeContext(ctx);
    p->v_readptr=0;

    stream->type = STREAMTYPE_BLANK;
    stream->flags = 0;
    stream->sector_size = 2048;
    stream->start_pos = 0;
    stream->end_pos = 0;

    stream->fill_buffer = fill_buffer;
    stream->close = close_s;
    stream->priv = opts;
    *file_format = DEMUXER_TYPE_BLANK;
    return 1;
}


const stream_info_t stream_info_blank = {
    "blank stream",
    "blank",
    "Otvos Attila",
    "",
    open_s,
    { "blank", NULL },
    &stream_opts,
    1 // Urls are an option string
};
