/*
 *  Share video frames through shared memory.
 *
 *  Copyright (C) 2007 Attila Ötvös <attila@onebithq.com>
 *
 *  MPlayer is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  MPlayer is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with MPlayer; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <inttypes.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#ifdef HAVE_SEMAPHORE
#include <sys/sem.h>
#endif
#include <signal.h>
#ifdef __MINGW32__
#define SIGUSR2 12
#endif

#include "mp_msg.h"
#include "stream.h"
#include "libmpdemux/demuxer.h"
#include "help_mp.h"
#include "m_option.h"
#include "m_struct.h"
#include "libavutil/common.h"
#include "mpbswap.h"
#include "shmem_common.h"
#include "stream_shmem.h"
#include "mp_msg.h"

//#undef HAVE_SEMAPHORE                    // disable semaphore support

#define WAIT_TIME          100             // 0.1 ms
#define SHMEM_SIZE         0x10000         // 64kB

#ifdef HAVE_SEMAPHORE

#define SEM_FILE          "/dev/null"      // file with create unique semID
#define SEM_NUM           1                // semaphores num
#define SEM_INIT          1                // semaphore init values
#define SEM_USLEEP        1000             // usleep
#define SEM_WAIT_TIME     20               // wait time = 20*1000 usec
#define SEM_SYNC_TIME     1000             // wait time = 1000*1000 usec
#define SEM_REQUIRE       1                // if can't open semaphore then exit
#define SEM_MEMBER        0                // semaphore num to read/write

#define SEM_OPEN          0                // reader - open exist semaphore
#define SEM_CREATE        1                // writer - create new semaphore

#if !defined(__GNU_LIBRARY__) || defined(_SEM_SEMUN_UNDEFINED)
// semun undefined in <sys/sem.h>
union semun
{
    int val;
    struct semid_ds *buf;
    unsigned short int *array;
    struct seminfo *__buf;
};
#endif
#endif

static int shmem_type = SHMEM_TYPE_PIPE;        // default type: pipe
static unsigned int shmem_size = SHMEM_SIZE;    // default size
// signal (use only write mode)
extern pid_t parent_pid;  // parend pid to send start (SIGUSR2) signal
extern int child_start;   // =1 if sent start signal

static struct stream_priv_s {
    char *ids;
    int mode;
    int id;
    int type;
    unsigned int size;
    int shmid;
    int semid;
    int flag;
    int ptr;
    int bytes;
    int started;
    char* data;
    char* shmem;
} stream_priv_dflts = {
    NULL,   // ids
    0,      // mode
    0,      // id
    0,      // type
    0,      // size
    0,      // shmid
    -1,            // semid
    0,      // flag
    0,            // ptr
    0,      // bytes
    0,      // started
    NULL,   // data
    NULL    // shmem
};


#define ST_OFF(f) M_ST_OFF(struct stream_priv_s,f)
/// URL definition
static m_option_t stream_opts_fields[] = {
    {"hostname", ST_OFF(ids), CONF_TYPE_STRING, 0, 0 ,0, NULL},
    { NULL, NULL, 0, 0, 0, 0,  NULL }
};
static struct m_struct_st stream_opts = {
    "shmem",
    sizeof(struct stream_priv_s),
    &stream_priv_dflts,
    stream_opts_fields
};

#ifdef HAVE_SEMAPHORE
/**
 * \brief get semaphore values
 * \param semid: semaphore ID
 * \param member: semaphore number
 * \return semaphore value
 */
static int semgetval(int semid, int member) {
    return semctl(semid, member, GETVAL, 0);
}

/**
 * \brief open semaphore
 * \param shmid: shared memory ID
 * \param writer: 0 - reader (open semaphore), 1 - writer (create semaphore)
 * \return semaphore ID, -1 on error
 */
static int semopen(int shmid, int writer) {
    int cpid, semid, i;
    key_t key;
    struct shmid_ds shmem_ds;
    union semun semopts;

    memset(&shmem_ds,0,sizeof(shmem_ds));
    cpid = 0;
    if (shmctl(shmid,IPC_STAT,&shmem_ds)==0)
        cpid = shmem_ds.shm_cpid;
    key = ftok(SEM_FILE,cpid);
    mp_msg(MSGT_OPEN,MSGL_V,"semaphore cpid: %d key: %X\n",cpid,key);
    if (writer) {
        if ((semid = semget(key, SEM_NUM, IPC_CREAT|IPC_EXCL|0666)) == -1) {
            if ((semid = semget(key, 0, 0666))!=-1)
                semctl(semid, IPC_RMID, 0);
            if ((semid = semget(key, SEM_NUM, IPC_CREAT|IPC_EXCL|0666)) == -1) {
                mp_msg(MSGT_OPEN,MSGL_ERR,
                       "Can't open shemaphore. (%s)\n",strerror(errno));
            return -1;
            }
        }
        semopts.val = SEM_INIT;
        for(i=0; i<SEM_NUM; i++)
            semctl(semid, i, SETVAL, semopts);
    } else {
        if ((semid = semget(key, 0, 0666))==-1) {
            mp_msg(MSGT_OPEN,MSGL_ERR,
                   "Can't open shemaphore. (%s)\n",strerror(errno));
            return -1;
        }
    }
    return semid;
}

/**
 * \brief check member is valid
 * \param semid: semaphore ID
 * \param member: semaphore number
 * \return 0 - on error, 1 - member is valid
 */
static int semismember(int semid, int member) {
    int membernum;
    union semun semopts;
    struct semid_ds semids;

    if (semid==-1) return 0;
    semopts.buf = &semids;
    semctl(semid, 0, IPC_STAT, semopts);
    membernum=semopts.buf->sem_nsems;
    if (member<0 || member>(membernum-1))
        return 1;
    return 0;
}

/**
 * \brief semaphore lock (no wait)
 * \param semid: semaphore ID
 * \param member: semaphore number
 * \return 0 - on error
 *         1 - OK
 *        -1 - semaphore locked
 */
static int semlock(int semid, int member) {
    struct sembuf sem_lock = { 0, -1, IPC_NOWAIT };

    if (semismember(semid, member)) return 0;
    if (!semgetval(semid, member)) return -1;
    sem_lock.sem_num = member;
    if ((semop(semid, &sem_lock, 1)) == -1) return 0;
    return 1;
}

/**
 * \brief semaphore lock (wait)
 * \param semid: semaphore ID
 * \param member: semaphore number
 * \return 0 - on error
 *         1 - OK
 */
static int semlockwait(int semid, int member) {
    struct sembuf sem_lock = { 0, -1, 0 };

    if (semismember(semid, member)) return 0;
    sem_lock.sem_num = member;
    if ((semop(semid, &sem_lock, 1)) == -1) return 0;
    return 1;
}

/**
 * \brief semaphore unlock
 * \param semid: semaphore ID
 * \param member: semaphore number
 */
static void semunlock(int semid, int member) {
    struct sembuf sem_unlock = { member, 1, IPC_NOWAIT };
    int semval;

    if (semismember(semid, member)) return;
    semval = semgetval(semid, member);
    if (semval == SEM_INIT) return;
    sem_unlock.sem_num = member;
    if ((semop(semid, &sem_unlock, 1)) == -1) return;
    return;
}
#define sem_lock(semid,member) semlock(semid,member)
#define sem_unlock(semid,member) semunlock(semid,member)
#else
#define sem_lock(semid,member) {}
#define sem_unlock(semid,member) {}
#endif

int set_shmem_size(int type, unsigned int size) {
    switch(type) {
    case SHMEM_TYPE_RAWDATA:
        if(!size)
            return 1;
        shmem_type=SHMEM_TYPE_RAWDATA;
        shmem_size=size+sizeof(shmem_data_t);
        return 0;
    case SHMEM_TYPE_PIPE:
        shmem_type=SHMEM_TYPE_PIPE;
        shmem_size=SHMEM_SIZE;
        return 0;
    }
    return 1;
}


/**
 * \brief get number of attach clients
 * \param semid: shmem ID
 * \return: clients number
 */
static int get_shmem_attach(int shmid) {
    struct shmid_ds smem_ds;

    smem_ds.shm_nattch=0;
    if (shmctl(shmid,IPC_STAT,&smem_ds)==0)
        return smem_ds.shm_nattch-1;
    return 0;
}

/**
 * \brief get memory size
 * \param semid: shmem ID
 * \return: memory size
 */
static int get_shmem_size(int shmid) {
    struct shmid_ds smem_ds;

    smem_ds.shm_segsz=0;
    if (shmctl(shmid,IPC_STAT,&smem_ds)==0)
        return smem_ds.shm_segsz;
    return 0;
}

static int fill_buffer_raw(stream_t *s, char* buffer, int max_len){
    struct stream_priv_s* p = (struct stream_priv_s*)s->priv;
    shmem_data_t *shmem_data;
    int len=-1;
    char *ptr;

    if (p->bytes) {
        len=FFMIN(p->bytes,max_len);
        ptr=p->data+p->ptr;
        memcpy(buffer,ptr,len);
        p->ptr+=len;
        p->bytes-=len;
        return len;
    }
#ifdef HAVE_SEMAPHORE
    if (p->semid!=-1) {
        if (!semlockwait(p->semid, SEM_MEMBER))
            return -1;
    }
#endif
    shmem_data=(shmem_data_t*)p->shmem;
    if (!get_shmem_attach(p->shmid)) {
        sem_unlock(p->semid, SEM_MEMBER);
        return -1;
    }
    p->bytes=FFMIN(p->size,shmem_data->size);
    p->ptr=0;
    memcpy(p->data,shmem_data->data,p->bytes);
    sem_unlock(p->semid, SEM_MEMBER);
    len=FFMIN(p->bytes,max_len);
    ptr=p->data+p->ptr;
    memcpy(buffer,ptr,len);
    p->ptr+=len;
    p->bytes-=len;
    return len;
}

static int fill_buffer_pipe(stream_t *s, char* buffer, int max_len){
    struct stream_priv_s* p = (struct stream_priv_s*)s->priv;
    shmem_data_t *shmem_data;
    int len=-1;
    char *ptr;

    if (p->bytes) {
        len=FFMIN(p->bytes,max_len);
        ptr=p->data+p->ptr;
        memcpy(buffer,ptr,len);
        p->ptr+=len;
        p->bytes-=len;
        return len;
    }
    sem_lock(p->semid, SEM_MEMBER);
    shmem_data=(shmem_data_t*)p->shmem;
    if(shmem_data->mplayerid[0]!='M' || shmem_data->mplayerid[1]!='P') {
        sem_unlock(p->semid, SEM_MEMBER);
        return -1;
    }
    if (stream_at_eof || is_shmem_eof(shmem_data->flag)) {
        sem_unlock(p->semid, SEM_MEMBER);
        return -1;
    }
    if (!get_shmem_attach(p->shmid)) {
        sem_unlock(p->semid, SEM_MEMBER);
        return -1;
    }
    while (!is_shmem_writed(shmem_data->flag)) {
        sem_unlock(p->semid, SEM_MEMBER);
        usleep(WAIT_TIME);
        sem_lock(p->semid, SEM_MEMBER);
        if (stream_at_eof || is_shmem_eof(shmem_data->flag)) {
            sem_unlock(p->semid, SEM_MEMBER);
            return -1;
        }
        if (!get_shmem_attach(p->shmid)) {
            sem_unlock(p->semid, SEM_MEMBER);
            return -1;
        }
    }
    p->bytes=FFMIN(p->size-sizeof(shmem_data_t),shmem_data->size);
    p->ptr=0;
    memcpy(p->data,shmem_data->data,p->bytes);
    shmem_data->flag&=~SHMEM_FLAG_WRITED;
    sem_unlock(p->semid, SEM_MEMBER);
    len=FFMIN(p->bytes,max_len);
    ptr=p->data+p->ptr;
    memcpy(buffer,ptr,len);
    p->ptr+=len;
    p->bytes-=len;
    return len;
}

static int fill_buffer(stream_t *s, char* buffer, int max_len){
    struct stream_priv_s* p = (struct stream_priv_s*)s->priv;
    switch(p->type) {
    case SHMEM_TYPE_RAWDATA:
        return fill_buffer_raw(s,buffer,max_len);
    case SHMEM_TYPE_PIPE:
        return fill_buffer_pipe(s,buffer,max_len);
    }
    return -1;
}


static int write_buffer_raw(stream_t *s, char* buffer, int len) {
    struct stream_priv_s* p = (struct stream_priv_s*)s->priv;
    shmem_data_t *shmem_data;
    unsigned int max_len=FFMIN(p->size,(unsigned int)len);
    int i, lockstatus;

#ifdef HAVE_SEMAPHORE
    if(p->semid != -1) {
        for(i=0;i<SEM_WAIT_TIME;i++) {
            lockstatus = semlock(p->semid, SEM_MEMBER);
            if(lockstatus !=-1) break;
            usleep(SEM_USLEEP);
        }
    if (lockstatus == -1) return len;              // skip frame
    if (lockstatus == 0) return -1;                // error
    }
#endif
    shmem_data=(shmem_data_t*)p->shmem;
    shmem_data->mplayerid[0]='M';
    shmem_data->mplayerid[1]='P';
    shmem_data->type=p->type;
    shmem_data->size=len;
    memcpy(shmem_data->data,buffer,max_len);
    sem_unlock(p->semid, SEM_MEMBER);
    return max_len;
}

static int write_buffer_pipe(stream_t *s, char* buffer, int len) {
    struct stream_priv_s* p = (struct stream_priv_s*)s->priv;
    shmem_data_t *shmem_data;
    int max_len=FFMIN(p->size-sizeof(shmem_data_t),(unsigned int)len);
    int attached=0;

    sem_lock(p->semid, SEM_MEMBER);
    shmem_data=(shmem_data_t*)p->shmem;
    if (stream_at_eof || is_shmem_eof(shmem_data->flag)) {
        sem_unlock(p->semid, SEM_MEMBER);
        return -1;
    }
    attached=get_shmem_attach(p->shmid);
    if (!attached && p->started) {
        sem_unlock(p->semid, SEM_MEMBER);
        return -1;
    }
    while (is_shmem_writed(shmem_data->flag)) {
        sem_unlock(p->semid, SEM_MEMBER);
        usleep(WAIT_TIME);
        sem_lock(p->semid, SEM_MEMBER);
        if (stream_at_eof || is_shmem_eof(shmem_data->flag)) {
            sem_unlock(p->semid, SEM_MEMBER);
            return -1;
        }
        if (p->started && !get_shmem_attach(p->shmid)) {
            sem_unlock(p->semid, SEM_MEMBER);
            return -1;
        }
    }
    shmem_data->mplayerid[0]='M';
    shmem_data->mplayerid[1]='P';
    shmem_data->type=p->type;
    shmem_data->size=max_len;
    memcpy(shmem_data->data,buffer,max_len);
    shmem_data->flag|=SHMEM_FLAG_WRITED;
    sem_unlock(p->semid, SEM_MEMBER);
    if (parent_pid && !child_start) {
        kill(parent_pid,SIGUSR2);
        child_start=1;
    }
    if (attached) p->started=1;
    return max_len;
}

static int write_buffer(stream_t *s, char* buffer, int len) {
    struct stream_priv_s* p = (struct stream_priv_s*)s->priv;
    switch(p->type) {
    case SHMEM_TYPE_RAWDATA:
        return write_buffer_raw(s,buffer,len);
    case SHMEM_TYPE_PIPE:
        return write_buffer_pipe(s,buffer,len);
    }
    return -1;
}

static int control(stream_t *stream,int cmd,void* arg) {
    struct stream_priv_s* p = (struct stream_priv_s*)stream->priv;
    switch(cmd) {
    case STREAM_CTRL_GET_TYPE:
        *(int*)arg=p->type;
        return 1;
    }
    return STREAM_UNSUPPORTED;
}

static void close_s(struct stream_st *s) {
    struct stream_priv_s* p = (struct stream_priv_s*)s->priv;

#ifdef HAVE_SEMAPHORE
    if(p->semid!=-1 && p->mode==STREAM_WRITE)
        semctl(p->semid, IPC_RMID, 0);
    p->semid=-1;
#endif
    shmdt(p->shmem);
    if(p->mode==STREAM_WRITE)
        shmctl(p->shmid, IPC_RMID, 0);
    p->shmem=NULL;
    if(p->data) free(p->data);
    m_struct_free(&stream_opts,p);
}


static int open_r(stream_t *stream, int mode, void* opts, int* file_format) {
    struct stream_priv_s* p = (struct stream_priv_s*)opts;
    shmem_data_t *shmem_data;

    if(0>(p->shmid=shmget(p->id,sizeof(shmem_data_t), 0666))) {
        mp_msg(MSGT_STREAM,MSGL_ERR,
               "Can't open shared memory: #%d. (%s)\n",p->id,strerror(errno));
        return STREAM_UNSUPPORTED;
    }
    if((p->shmem=shmat(p->shmid,0,0))==(char*)-1) {
        shmdt(p->shmem);
        mp_msg(MSGT_STREAM,MSGL_ERR,
               "Can't attach shared memory: #%d. (%s)\n",p->id,strerror(errno));
        return STREAM_UNSUPPORTED;
    }
#ifdef HAVE_SEMAPHORE
    p->semid = semopen(p->shmid,SEM_OPEN);
    if (p->semid == -1 && SEM_REQUIRE) {
        shmdt(p->shmem);
        return STREAM_UNSUPPORTED;
    }
    if (p->semid != -1) {
        if (!semlockwait(p->semid, SEM_MEMBER)) {
            semctl(p->semid, IPC_RMID, 0);
            shmdt(p->shmem);
            return STREAM_UNSUPPORTED;
        }
    }
#endif
    shmem_data=(shmem_data_t*)p->shmem;
    if(shmem_data->mplayerid[0]!='M' || shmem_data->mplayerid[1]!='P') {
        sem_unlock(p->semid, SEM_MEMBER);
        shmdt(p->shmem);
        return STREAM_UNSUPPORTED;
    }
    p->size=get_shmem_size(p->shmid);
    switch (shmem_data->type) {
    case SHMEM_TYPE_RAWDATA:
    case SHMEM_TYPE_PIPE:
        p->type=shmem_data->type;
        break;
    default:
        mp_msg(MSGT_STREAM,MSGL_ERR,
               "Unknown type %d in shared memory: #%d.\n",
               shmem_data->type,p->id);
        sem_unlock(p->semid, SEM_MEMBER);
        shmdt(p->shmem);
        return STREAM_UNSUPPORTED;
    }
    shmdt(p->shmem);
    if(0>(p->shmid=shmget(p->id,p->size, 0666))) {
        sem_unlock(p->semid, SEM_MEMBER);
        mp_msg(MSGT_STREAM,MSGL_ERR,
               "Can't reopen shared memory: #%d. (%s)\n",p->id,strerror(errno));
        return STREAM_UNSUPPORTED;
    }
    if((p->shmem=shmat(p->shmid,0,0))==(char*)-1) {
        sem_unlock(p->semid, SEM_MEMBER);
        shmdt(p->shmem);
        mp_msg(MSGT_STREAM,MSGL_ERR,
               "Can't attach shared memory: #%d. (%s)\n",p->id,strerror(errno));
        return STREAM_UNSUPPORTED;
    }
    p->data=calloc(1,p->size);
    sem_unlock(p->semid, SEM_MEMBER);
    p->mode=mode;
    stream->type = STREAMTYPE_SHMEM;
    stream->flags = 0;
    stream->sector_size = 2048;
    stream->start_pos = 0;
    stream->end_pos = 0;

    stream->fill_buffer = fill_buffer;
    stream->control = control;
    stream->close = close_s;
    stream->priv = opts;

    *file_format = DEMUXER_TYPE_SHMEM;
    return STREAM_OK;
}

static int open_w(stream_t *stream, int mode, void* opts, int* file_format) {
    struct stream_priv_s* p = (struct stream_priv_s*)opts;
    shmem_data_t *shmem_data;

    p->size=shmem_size;
    p->type=shmem_type;
    shmem_type = SHMEM_TYPE_PIPE;
    shmem_size = SHMEM_SIZE;

    if(0>(p->shmid=shmget(p->id, p->size, 0666 | IPC_CREAT))) {
        if(0>(p->shmid=shmget(p->id, sizeof(shmem_data_t), 0666))) {
            mp_msg(MSGT_STREAM,MSGL_ERR,
                   "Can't create shared memory: #%d. (%s)\n",
                   p->id,strerror(errno));
            return STREAM_UNSUPPORTED;
        }
        if(get_shmem_attach(p->shmid)) {
            mp_msg(MSGT_STREAM,MSGL_ERR,
                   "Shared memory is busy. Can't create.: #%d. (%s)\n",
                   p->id,strerror(errno));
            return STREAM_UNSUPPORTED;
        }
        shmctl(p->shmid, IPC_RMID, 0);          // Delete if unused
        if(0>(p->shmid=shmget(p->id, p->size, 0666 | IPC_CREAT))) {
            mp_msg(MSGT_STREAM,MSGL_ERR,
                   "Can't create shared memory: #%d. (%s)\n",
                   p->id,strerror(errno));
            return STREAM_UNSUPPORTED;
        }
    }
    if ((p->shmem=shmat(p->shmid,0,0))==(char*)-1) {
        mp_msg(MSGT_STREAM,MSGL_ERR,
               "Can't attach shared memory: #%d (%s)\n",p->id,strerror(errno));
        return STREAM_UNSUPPORTED;
    }
#ifdef HAVE_SEMAPHORE
    p->semid = semopen(p->shmid,SEM_CREATE);
    if (p->semid == -1 && SEM_REQUIRE) {
        shmdt(p->shmem);
        shmctl(p->shmid, IPC_RMID, 0);
        mp_msg(MSGT_STREAM,MSGL_ERR,
               "Can't create semapore: #%d (%s)\n",p->id,strerror(errno));
        return STREAM_UNSUPPORTED;
    }
    if (p->semid != -1) {
        if (!semlockwait(p->semid, SEM_MEMBER)) {
            semctl(p->semid, IPC_RMID, 0);
            shmdt(p->shmem);
            shmctl(p->shmid, IPC_RMID, 0);
            mp_msg(MSGT_STREAM,MSGL_ERR,
                   "Can't lock semapore: #%d (%s)\n",p->id,strerror(errno));
            return STREAM_UNSUPPORTED;
        }
    }
#endif
    memset(p->shmem,0,p->size);
    shmem_data=(shmem_data_t*)p->shmem;
    shmem_data->mplayerid[0]='M';
    shmem_data->mplayerid[1]='P';
    shmem_data->type=p->type;
    shmem_data->size=0;
    p->data=calloc(1,p->size);
    sem_unlock(p->semid, SEM_MEMBER);
    p->mode=mode;
    stream->type = STREAMTYPE_SHMEM;
    stream->flags = 0;
    stream->sector_size = 2048;
    stream->start_pos = 0;
    stream->end_pos = 0;

    stream->write_buffer = write_buffer;
    stream->control = control;
    stream->close = close_s;
    stream->priv = opts;
    *file_format=0;
    return STREAM_OK;
}

static int open_s(stream_t *stream,int mode, void* opts, int* file_format) {
    struct stream_priv_s* p = (struct stream_priv_s*)opts;
    int id = 0;

    if (p->ids)
        id = atoi(p->ids);
    if (!id)
        return STREAM_UNSUPPORTED;
    p->id=id;
    if (mode==STREAM_WRITE)
        return open_w(stream,mode,opts,file_format);
    if (mode==STREAM_READ)
        return open_r(stream,mode,opts,file_format);
    return STREAM_UNSUPPORTED;
}

stream_info_t stream_info_shmem = {
    "Shared memory stream",
    "shmem",
    "Otvos Attila",
    "",
    open_s,
    { "shmem",NULL },
    &stream_opts,
    1 // Url is an option string
};
